"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Helper functions relating to tab separated text files. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


import csv
from duHast.Utilities.Objects.result import Result
from duHast.Utilities.files_base_write import write_report_data
from duHast.Utilities.files_base_read import (
    get_first_row_in_column_based_text_file,
    read_column_based_text_file,
)
from duHast.Utilities.files_base_combine import (
    combine_files,
    combine_files_header_independent,
    append_to_file,
)
from duHast.Utilities.files_get import get_files_single_directory


def append_tab_separated_file(
    source_file, append_file, ignore_first_row=False, quoting=csv.QUOTE_MINIMAL
):
    """
    Function to append the content of a tab separated file to another tab separated file.

    :param source_file: The fully qualified file path of the source file.
    :type source_file: str
    :param append_file: The fully qualified file path of the file to append.
    :type append_file: str
    :param ignore_first_row: Flag to ignore the first row of the append file. Defaults to False.
    :type ignore_first_row: bool, optional
    :param quoting: Quoting style used by the csv writer. Defaults to csv.QUOTE_MINIMAL. Options are csv.QUOTE_ALL, csv.QUOTE_MINIMAL, csv.QUOTE_NONNUMERIC, csv.QUOTE_NONE
    :type quoting: int, optional

    :return:
        Result class instance.

        - result.status (bool) True if file was appended without an exception, otherwise False.
        - result.message contains log messages.
        - result.result will be an empty list.

        On exception:

        - result.status (bool) will be False.
        - result.message will contain exception message.
    :rtype: :class:`.Result`
    """

    # use base function to append the file
    return_value = append_to_file(
        source_file=source_file,
        append_file=append_file,
        ignore_first_row=ignore_first_row,
        delimiter="\t",
        quoting=quoting,
    )

    return return_value


def combine_tab_separated_files_header_independent(
    folder_path,
    file_prefix="",
    file_suffix="",
    file_extension=".txt",
    output_file_name="result.txt",
    overwrite_existing=False,
):
    """
    Function to combine multiple tab separated files into a single tab separated file combining all headers.

    :param folder_path: The fully qualified folder path containing the tab separated files.
    :type folder_path: str
    :param file_prefix: The prefix of the tab separated files to be combined. Defaults to "".
    :type file_prefix: str, optional
    :param file_suffix: The suffix of the tab separated files to be combined. Defaults to "".
    :type file_suffix: str, optional
    :param file_extension: The extension of the tab separated files to be combined. Defaults to ".txt".
    :type file_extension: str, optional
    :param output_file_name: The name of the output file. Defaults to "result.txt".
    :type output_file_name: str, optional
    :param overwrite_existing: Flag to overwrite the existing output file. Defaults to False.
    :type overwrite_existing: bool, optional

    :return:
        Result class instance.

        - result.status (bool) True if files were combined without an exception, otherwise False.
        - result.message contains log messages.
        - result.result will be an complete list of all rows appended.

        On exception:

        - result.status (bool) will be False.
        - result.message will contain exception message.
    :rtype: :class:`.Result`
    """

    # use base function to combine the files
    return_value = combine_files_header_independent(
        folder_path=folder_path,
        file_prefix=file_prefix,
        file_suffix=file_suffix,
        file_extension=file_extension,
        output_file_name=output_file_name,
        overwrite_existing=overwrite_existing,
        delimiter="\t",
    )

    return return_value


def combine_tab_separated_files(
    folder_path,
    file_prefix="",
    file_suffix="",
    file_extension=".txt",
    output_file_name="result.txt",
    file_getter=get_files_single_directory,
    quoting=csv.QUOTE_MINIMAL,
):
    """
    Function to combine multiple tab separated files into a single tab separated file.

    Assumes all files have the same header. (number of columns)

    :param folder_path: The fully qualified folder path containing the tab separated files.
    :type folder_path: str
    :param file_prefix: The prefix of the tab separated files to be combined. Defaults to "".
    :type file_prefix: str, optional
    :param file_suffix: The suffix of the tab separated files to be combined. Defaults to "".
    :type file_suffix: str, optional
    :param file_extension: The extension of the tab separated files to be combined. Defaults to ".txt".
    :type file_extension: str, optional
    :param output_file_name: The name of the output file. Defaults to "result.txt".
    :type output_file_name: str, optional
    :param file_getter: Function to get the files in the folder. Defaults to get_files_single_directory.
    :type file_getter: function, optional
    :param quoting: Quoting style used by the tab separated writer. Defaults to csv.QUOTE_MINIMAL. Options are csv.QUOTE_ALL, csv.QUOTE_MINIMAL, csv.QUOTE_NONNUMERIC, csv.QUOTE_NONE
    :type quoting: int, optional

    :return:
        Result class instance.

        - result.status (bool) True if files were combined without an exception, otherwise False.
        - result.message contains log messages.
        - result.result will be an empty list.

        On exception:

        - result.status (bool) will be False.
        - result.message will contain exception message.
    :rtype: :class:`.Result`
    """

    # use base function to combine the files
    return_value = combine_files(
        folder_path=folder_path,
        file_prefix=file_prefix,
        file_suffix=file_suffix,
        file_extension=file_extension,
        output_file_name=output_file_name,
        file_getter=file_getter,
        delimiter="\t",
        quoting=quoting,
    )

    return return_value


def write_report_data_as_tab_separated_file(
    file_name,
    header,
    data,
    write_type="w",
    enforce_ascii=False,
    encoding="utf-8",
    bom=None,
    quoting=csv.QUOTE_NONE,
):
    """
    Function writing out report information in tab separated format.

    :param file_name: The reports fully qualified file path.
    :type file_name: str
    :param header: list of column headers
    :type header: list of str
    :param data: list of list of strings representing row data
    :type data: [[str,str,..]]
    :param write_type: Flag indicating whether existing report file is to be overwritten 'w' or appended to 'a', defaults to 'w'
    :type write_type: str, optional
    :param enforce_ascii: Flag to enforce ASCII encoding on data. If True, data will be encoded to ASCII. Defaults to False.
    :type enforce_ascii: bool, optional
    :param encoding: Encoding used to write the file. Defaults to 'utf-8'.
    :type encoding: str, optional
    :param bom: the byte order mark, Default is None (none will be written). BOM: "utf-16" = , "utf-16-le" = ,  utf-8 =
    :type bom: str, default is NoneType
    :param quoting: Quoting style used by the csv writer. Defaults to csv.QUOTE_NONE. Options are csv.QUOTE_ALL, csv.QUOTE_MINIMAL, csv.QUOTE_NONNUMERIC, csv.QUOTE_NONE
    :type quoting: int, optional

    :return: A Result object, with the result attribute set to True if the file was written successfully, False otherwise.
    :rtype: :class:`.Result`
    """

    write_result = write_report_data(
        file_name=file_name,
        header=header,
        data=data,
        write_type=write_type,
        enforce_ascii=enforce_ascii,
        encoding=encoding,
        quoting=quoting,
        bom=bom,
        delimiter="\t",
    )
    return write_result


def get_first_row_in_tab_separated_file(file_path):
    """
    Reads the first line of a tab separated text file and returns it as a list of strings

    :param file_path: The fully qualified file path.
    :type file_path: str
    :return:
        Result class instance.

        - result.status (bool) True if file was read without an exception, otherwise False.
        - result.message contains log messages.
        - result.result will contain a list of list of strings representing the first row.

        On exception:

        - result.status (bool) will be False.
        - result.message will contain exception message.
    :rtype: :class:`.Result`
    """

    return_value = Result()
    return_value = get_first_row_in_column_based_text_file(
        file_path=file_path, delimiter="\t"
    )
    return return_value


def read_tab_separated_file(file_path, increase_max_field_size_limit=False):
    """
    Read a tab separated file into a list of rows, where each row is another list.

    :param file_path: The fully qualified file path to the tab separated text file.
    :type file_path: str
    :param increase_max_field_size_limit: Flag to increase the max field size limit. Defaults to False.
    :type increase_max_field_size_limit: bool, optional

    :return:
        Result class instance.

        - result.status (bool) True if file was read without an exception, otherwise False.
        - result.message contains log messages.
        - result.result will contain a list of list of strings representing the data in each row.

        On exception:

        - result.status (bool) will be False.
        - result.message will contain exception message.
    :rtype: :class:`.Result`
    """

    # read with encoding enabled
    read_result = read_column_based_text_file(
        file_path=file_path,
        increase_max_field_size_limit=increase_max_field_size_limit,
        delimiter="\t",
    )

    return read_result



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit railings. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr
import System

from duHast.Revit.Common import common as com
from duHast.Revit.Family import family_utils as rFam

# import Autodesk
import Autodesk.Revit.DB as rdb
import Autodesk.Revit.DB.Architecture as rdbA

from duHast.Revit.Railings.Utility.railing_categories import RAILING_CATEGORY_FILTER
from duHast.Revit.Railings.Utility import railings_filter as rRailFilter

# --------------------------------------------- utility functions ------------------


def get_all_railing_types_by_category(doc):
    """
    Gets a filtered element collector of all Railing types in the model.

    Collector will include types of:
    - Top Rail
    - Rail support
    - Hand rail
    - Rail termination
    - Railing Systems
    - In place families or loaded families

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector of railing related types
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = rRailFilter._get_all_railing_types_by_category(doc)
    return collector


def get_all_railing_types_by_category_excl_in_place(doc):
    """
    Gets a filtered element collector of all Railing types in the model.

    Collector will include types of:
    - Top Rail
    - Rail support
    - Hand rail
    - Rail termination
    - Railing Systems
    - loaded families

    Will exclude any inplace families.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A list railing related types
    :rtype: list of types
    """

    multi_cat_filter = rdb.ElementMulticategoryFilter(RAILING_CATEGORY_FILTER)
    collector = (
        rdb.FilteredElementCollector(doc)
        .WherePasses(multi_cat_filter)
        .WhereElementIsElementType()
    )
    elements = []
    for c in collector:
        if c.GetType() != rdb.FamilySymbol:
            elements.append(c)
    return elements


def get_railing_types_by_class(doc):
    """
    Gets a filtered element collector of all Railing types in the model:

    Collector will include types of:
    - Railing

    It will therefore not return any top rail or hand rail or in place family types.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector of railing types
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = rRailFilter._get_railing_types_by_class(doc)
    return collector


# -------------------------------- none in place Railing types -------------------------------------------------------


def get_all_railing_instances_by_category(doc):
    """
    Gets all Railing elements placed in model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector of railing instances.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    multi_cat_filter = rdb.ElementMulticategoryFilter(RAILING_CATEGORY_FILTER)
    return (
        rdb.FilteredElementCollector(doc)
        .WherePasses(multi_cat_filter)
        .WhereElementIsNotElementType()
    )


def get_all_railing_instances_by_class(doc):
    """
    Gets all Railing elements placed in model. Ignores any in place families.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector of railing instances.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    return (
        rdb.FilteredElementCollector(doc)
        .OfClass(rdbA.Railing)
        .WhereElementIsNotElementType()
    )


def get_all_railing_type_ids_by_category(doc):
    """
    Gets all railing element type ids available in model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of element ids of railing types.
    :rtype: List Autodesk.Revit.DB.ElementId
    """

    ids = []
    col_cat = get_all_railing_types_by_category(doc)
    ids = com.get_ids_from_element_collector(col_cat)
    return ids


def get_all_railing_type_ids_by_class(doc):
    """
    Gets all railing element type ids available in model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of element ids of railing types.
    :rtype: List Autodesk.Revit.DB.ElementId
    """

    ids = []
    col_class = get_railing_types_by_class(doc)
    ids = com.get_ids_from_element_collector(col_class)
    return ids


def get_all_railing_type_ids_by_class_and_category(doc):
    """
    Gets all Railing element types available in model excluding in place types.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of element ids of railing types.
    :rtype: list Autodesk.Revit.DB.ElementId
    """

    ids = []
    col_class = get_railing_types_by_class(doc)
    ids_class = com.get_ids_from_element_collector(col_class)
    col_cat = get_all_railing_types_by_category_excl_in_place(doc)
    ids_cat = com.get_ids_from_element_collector(col_cat)
    for id_class in ids_class:
        if id_class not in ids:
            ids.append(id_class)
    for id_cat in ids_cat:
        if id_cat not in ids:
            ids.append(id_cat)
    return ids


# -------------------------------- In place Railing types -------------------------------------------------------


def get_in_place_railing_family_instances(doc):
    """
    Gets all instances of in place families of category Railing in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector of railing instances.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    filter = rdb.ElementMulticategoryFilter(RAILING_CATEGORY_FILTER)
    return (
        rdb.FilteredElementCollector(doc)
        .OfClass(rdb.FamilyInstance)
        .WherePasses(filter)
    )


def get_in_place_railing_type_ids_in_model(doc):
    """
    Gets type ids off all available in place families of category Railing.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of element ids of in place railing types.
    :rtype: list Autodesk.Revit.DB.ElementId
    """

    ids = []
    for cat in RAILING_CATEGORY_FILTER:
        ids_by_cat = rFam.get_all_in_place_type_ids_in_model_of_category(doc, cat)
        if len(ids_by_cat) > 0:
            ids = ids + ids_by_cat
    return ids



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a warning guids of specific Revit warnings. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#


# area line is slightly off axis:
AREA_LINE_OFF_AXIS = "77535238-e93a-46c4-9822-97072116aefc"

# area separation lines overlap
AREA_SEPARATION_LINES_OVERLAP = "6891f0e6-5858-4e2f-bb34-801e0b87f60d"

# room tag outside room
ROOM_TAG_OUTSIDE_ROOM = "4f0bba25-e17f-480a-a763-d97d184be18a"

# room separation line overlaps
# Highlighted room separation lines overlap. One of them may be ignored when Revit finds room boundaries. Delete one of the lines.
ROOM_SEPARATION_LINES_OVERLAP = "374396c0-984d-4f72-a081-30ab7dacb66d"

# a room separation line and wall overlap
# A wall and a room separation line overlap. One of them may be ignored when Revit finds room boundaries. Shorten or delete the room separation line to remove the overlap.
ROOM_AND_WALL_SEPARATION_LINE_OVERLAP = "f7b3a015-c3eb-4a3f-b345-c474ec07d43f"

# Room separation line is slightly off axis and may cause inaccuracies.
ROOM_LINE_OFF_AXIS = "f657364a-e0b7-46aa-8c17-edd8e59683b9"

# Room is not in a properly enclosed region
ROOM_NOT_ENCLOSED = "ce3275c6-1c51-402e-8de3-df3a3d566f5c"

# duplicate mark value
DUPLICATE_MARK_VALUE = "6e1efefe-c8e0-483d-8482-150b9f1da21a"

# Element owned by others
# Can't edit the element until \"somebody\" re-saves the element to central and relinquishes it and you Reload Latest.
ELEMENTS_OWNED_BY_OTHER = "3b7fcaec-c01e-4c2e-819f-67ddd102ce1f"

# Elements have duplicate "Number" values.
ELEMENTS_DUPLICATE_NUMBER_VALUES = "6e1efefe-c8e0-483d-8482-150b9f1da21a"

# There are identical instances in the same place. This will result in double counting in schedules. You can tab-select one of the overlapping elements to exclude it from the group instance.
IDENTICAL_INSTANCES_IN_SAME_PLACE = "b4176cef-6086-45a8-a066-c3fd424c9412"

# Highlighted lines overlap. Lines may not form closed loops. You can tab-select one of the overlapping elements to exclude it from the group instance.
LINES_OVERLAP = "6891f0e6-5858-4e2f-bb34-801e0b87f60d"

# Line in Sketch is slightly off axis and may cause inaccuracies.
LINE_IN_SKETCH_OFF_AXIS = "ec905b6a-064b-48e1-a6c7-51e2cf092490"

# Line in Massing Sketch is slightly off axis and may cause inaccuracies.
LINE_IN_MASSING_SKETCH_OFF_AXIS = "e4b4847a-1cbc-4283-854e-85431403a2a0"

# Line is slightly off axis and may cause inaccuracies.
LINE_OFF_AXIS = "7b8f75b6-a3f2-48a4-8b5d-d767c9a20b32"

# Wall is slightly off axis and may cause inaccuracies.
WALL_OFF_AXIS = "fb1d4583-6b62-4ccb-9cb8-7c2e849b4d43"

# Ref Plane is slightly off axis and may cause inaccuracies.
REF_PLANE_OFF_AXIS = "ce961443-d03b-4cda-94c9-6b9beab9d2a6"

# Curve-Based Family is slightly off axis and may cause inaccuracies.
FAMILY_CURVE_BASED_OFF_AXIS = "e9b42850-030f-4909-990b-cf7cddf165a6"



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit area separation lines reports functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Areas.areas import get_area_schemes
from duHast.Revit.Areas.area_lines import get_area_lines_by_scheme_and_level_name
from duHast.Revit.Levels.levels import get_levels_list_ascending
from duHast.Revit.Warnings.warning_guids import AREA_LINE_OFF_AXIS, AREA_SEPARATION_LINES_OVERLAP
from duHast.Revit.Warnings.warnings import get_unique_warnings_elements_by_guid

from Autodesk.Revit.DB import Element


def area_lines_by_schemes_and_by_levels (doc):
    """
    Reports all area separation line by area schemes and level they belong too. 

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: dictionary where key is the area scheme name and value is a list of are line ids, or empty list if no area lines are associated with the scheme
    :rtype: {str: Autodesk.Revit.DB.ElementId]}
    """


    line_by_area_scheme_and_level = {}

    all_area_schemes = get_area_schemes(doc=doc)
    levels_ascending = get_levels_list_ascending(doc=doc)

    # loop over all area schemes in the model
    for area_scheme in all_area_schemes:
        
        # get the scheme name
        area_scheme_name = Element.Name.GetValue(area_scheme)
        
        # set up a dictionary containing level name as key and list of area lines as values
        by_level_name = {}

        # loop over levels
        for level in levels_ascending:
            # get the level name
            level_name = Element.Name.GetValue(level)

            # get lines by level
            area_lines_by_scheme_and_level = get_area_lines_by_scheme_and_level_name(doc=doc, scheme_name=area_scheme_name, level_name=level_name)

            # check for None
            if area_lines_by_scheme_and_level == None:
                by_level_name[level_name] = []
            else:
                by_level_name[level_name]=area_lines_by_scheme_and_level
        
        # store room separation lines by level in by scheme dictionary
        line_by_area_scheme_and_level[area_scheme_name] = by_level_name
    
    return line_by_area_scheme_and_level


def area_lines_with_warnings_by_schemes_and_by_levels(doc):
    """
    Reports all area separation line with warnings by area schemes and level they belong too. 

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: dictionary where key is the area scheme name and value is a list of are line ids, or empty list if no area lines are associated with the scheme
    :rtype: {str: Autodesk.Revit.DB.ElementId]}
    """

    # get area lines by scheme name and level name
    line_by_area_scheme_and_level = area_lines_by_schemes_and_by_levels (doc= doc)

    # check for empty and return if so
    if(len(line_by_area_scheme_and_level)==0):
        return line_by_area_scheme_and_level

    # get all area lines with warnings attached to them
    area_line_ids_of_axis = get_unique_warnings_elements_by_guid(doc=doc, guid= AREA_LINE_OFF_AXIS)

    # get all area lines overlapping
    area_lines_ids_overlapping = get_unique_warnings_elements_by_guid(doc=doc, guid= AREA_SEPARATION_LINES_OVERLAP)

    # combine lines with warnings lists
    area_lines_with_warnings = area_line_ids_of_axis + area_lines_ids_overlapping

    # check for empty and return if so
    if len(area_lines_with_warnings)==0:
        return {}
    
    # loop over dictionary and find lines with warnings
    for scheme_name in line_by_area_scheme_and_level:

        for level_name, separation_lines in line_by_area_scheme_and_level[scheme_name].items():
            lines_with_warnings = []
            for sep_line in separation_lines:
                if sep_line.Id in area_lines_with_warnings:
                    lines_with_warnings.append(sep_line)
            # replace previous line id list with line ids with warnigns associated
            line_by_area_scheme_and_level[scheme_name][level_name]=lines_with_warnings
    

    return line_by_area_scheme_and_level


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Contains file names used by model health report.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

#: header for each report file
LOG_FILE_HEADER = ["HOSTFILE", "REPORT", "DATE", "TIME", "VALUE"]


#: report file name suffix : date last updated
PARAM_ACTIONS_FILENAME_DATE_LAST_UPDATED = "_DateLastUpdated"
#: report file name suffix : worksets
PARAM_ACTIONS_FILENAME_NO_OF_WORKSETS = "_NumberOfWorksets"
#: report file name suffix : file size
PARAM_ACTIONS_FILENAME_FILE_SIZE = "_FileSize"
#: report file name suffix : warnings
PARAM_ACTIONS_FILENAME_NO_OF_WARNINGS = "_NumberOfWarnings"
#: report file name suffix : design sets
PARAM_ACTIONS_FILENAME_NO_OF_DESIGN_SETS = "_NumberOfDesignSets"
#: report file name suffix : design options
PARAM_ACTIONS_FILENAME_NO_OF_DESIGN_OPTIONS = "_NumberOfDesignOptions"
#: report file name suffix : sheet
PARAM_ACTIONS_FILENAME_NO_OF_SHEETS = "_NumberOfSheets"
#: report file name suffix : views
PARAM_ACTIONS_FILENAME_NO_OF_VIEWS = "_NumberOfViews"
#: report file name suffix : views not placed
PARAM_ACTIONS_FILENAME_NO_OF_VIEWS_NOT_PLACED = "_NumberOfViewsNotPlaced"
#: report file name suffix : view filters
PARAM_ACTIONS_FILENAME_NO_OF_VIEW_FILTERS = "_NumberOfViewFilters"
#: report file name suffix : view filters unused
PARAM_ACTIONS_FILENAME_NO_OF_VIEW_FILTERS_UNUSED = "_NumberOfViewFiltersUnused"
#: report file name suffix : text styles
PARAM_ACTIONS_FILENAME_NO_OF_TEXT_STYLES = "_NumberOfTextStyles"
#: report file name suffix : view filters unused
PARAM_ACTIONS_FILENAME_NO_OF_TEXT_STYLES_UNUSED = "_NumberOfTextStylesUnused"
#: report file name suffix : dimension styles
PARAM_ACTIONS_FILENAME_NO_OF_DIMENSION_STYLES = "_NumberOfDimStyles"
#: report file name suffix : dimension unused
PARAM_ACTIONS_FILENAME_NO_OF_DIMENSION_STYLES_UNUSED = "_NumberOfDimStylesUnused"
#: report file name suffix : arrow head styles
PARAM_ACTIONS_FILENAME_NO_OF_ARROW_HEAD_STYLES = "_NumberOfArrowHeadStyles"
#: report file name suffix : arrow head styles unused
PARAM_ACTIONS_FILENAME_NO_OF_ARROW_HEAD_STYLES_UNUSED = "_NumberOfArrowHeadStylesUnused"
#: report file name suffix : view templates
PARAM_ACTIONS_FILENAME_NO_OF_VIEW_TEMPLATES = "_NumberOfViewTemplates"
#: report file name suffix : view template unused
PARAM_ACTIONS_FILENAME_NO_OF_VIEW_TEMPLATES_UNUSED = "_NumberOfViewTemplatesUnused"
#: report file name suffix : line styles
PARAM_ACTIONS_FILENAME_NO_OF_LINE_STYLES = "_NumberOfLineStyles"
#: report file name suffix : line patterns
PARAM_ACTIONS_FILENAME_NO_OF_LINE_PATTERNS = "_NumberOfLinePatterns"
#: report file name suffix : fill patterns
PARAM_ACTIONS_FILENAME_NO_OF_FILL_PATTERNS = "_NumberOfFillPatterns"
#: report file name suffix : CAD imports
PARAM_ACTIONS_FILENAME_NO_OF_CAD_IMPORTS = "_NumberOfCadImports"
#: report file name suffix : CAD links in model
PARAM_ACTIONS_FILENAME_NO_OF_CAD_LINKS_MODEL = "_NumberOfCadLinksModel"
#: report file name suffix : CAD links in view
PARAM_ACTIONS_FILENAME_NO_OF_CAD_LINKS_VIEW = "_NumberOfCadLinksView"
#: report file name suffix : image imports
PARAM_ACTIONS_FILENAME_NO_OF_IMAGE_IMPORTS = "_NumberOfImageImports"
#: report file name suffix : image links
PARAM_ACTIONS_FILENAME_NO_OF_IMAGE_LINKS = "_NumberOfImageLinks"
#: report file name suffix : families in model
PARAM_ACTIONS_FILENAME_NO_OF_FAMILIES = "_NumberOfFamilies"
#: report file name suffix : in place families in model
PARAM_ACTIONS_FILENAME_NO_OF_FAMILIES_IN_PLACE = "_NumberOfInPlaceFamilies"
#: report file name suffix : model groups
PARAM_ACTIONS_FILENAME_NO_OF_MODEL_GROUPS = "_NumberOfModelGroups"
#: report file name suffix : model groups unplaced
PARAM_ACTIONS_FILENAME_NO_OF_MODEL_GROUPS_UNPLACED = "_NumberOfUnplacedModelGroups"
#: report file name suffix : detail groups
PARAM_ACTIONS_FILENAME_NO_OF_DETAIL_GROUPS = "_NumberOfDetailGroups"
#: report file name suffix : detail groups unplaced
PARAM_ACTIONS_FILENAME_NO_OF_DETAIL_GROUPS_UNPLACED = "_NumberOfUnplacedDetailGroups"
#: report file name suffix : rooms
PARAM_ACTIONS_FILENAME_NO_OF_ROOMS = "_NumberOfRooms"
#: report file name suffix : unplaced rooms
PARAM_ACTIONS_FILENAME_NO_OF_ROOMS_UNPLACED = "_NumberOfUnplacedRooms"
#: report file name suffix : not enclosed rooms
PARAM_ACTIONS_FILENAME_NO_OF_ROOMS_UNENCLOSED = "_NumberOfNotEnclosedRooms"
#: report file name suffix : redundant rooms
PARAM_ACTIONS_FILENAME_NO_OF_ROOMS_REDUNDANT = "_NumberOfRedundantRooms"
#: report file name suffix : filled regions
PARAM_ACTIONS_FILENAME_NO_OF_FILLED_REGIONS = "_NumberOfRegions"

# combined log file for all reports
PARAM_ACTIONS_FILENAME_MOTHER = "_AllLogs"


#: list of report file name extensions
PARAM_ACTIONS_FILENAMES = {
    PARAM_ACTIONS_FILENAME_NO_OF_WORKSETS,
    PARAM_ACTIONS_FILENAME_FILE_SIZE,
    PARAM_ACTIONS_FILENAME_NO_OF_WARNINGS,
    PARAM_ACTIONS_FILENAME_NO_OF_DESIGN_SETS,
    PARAM_ACTIONS_FILENAME_NO_OF_DESIGN_OPTIONS,
    PARAM_ACTIONS_FILENAME_NO_OF_SHEETS,
    PARAM_ACTIONS_FILENAME_NO_OF_VIEWS,
    PARAM_ACTIONS_FILENAME_NO_OF_VIEWS_NOT_PLACED,
    PARAM_ACTIONS_FILENAME_NO_OF_VIEW_FILTERS,
    PARAM_ACTIONS_FILENAME_NO_OF_VIEW_FILTERS_UNUSED,
    PARAM_ACTIONS_FILENAME_NO_OF_VIEW_TEMPLATES,
    PARAM_ACTIONS_FILENAME_NO_OF_VIEW_TEMPLATES_UNUSED,
    PARAM_ACTIONS_FILENAME_NO_OF_LINE_STYLES,
    PARAM_ACTIONS_FILENAME_NO_OF_LINE_PATTERNS,
    PARAM_ACTIONS_FILENAME_NO_OF_FILL_PATTERNS,
    PARAM_ACTIONS_FILENAME_NO_OF_CAD_IMPORTS,
    PARAM_ACTIONS_FILENAME_NO_OF_CAD_LINKS_MODEL,
    PARAM_ACTIONS_FILENAME_NO_OF_CAD_LINKS_VIEW,
    PARAM_ACTIONS_FILENAME_NO_OF_IMAGE_IMPORTS,
    PARAM_ACTIONS_FILENAME_NO_OF_IMAGE_LINKS,
    PARAM_ACTIONS_FILENAME_NO_OF_FAMILIES,
    PARAM_ACTIONS_FILENAME_NO_OF_FAMILIES_IN_PLACE,
    PARAM_ACTIONS_FILENAME_NO_OF_MODEL_GROUPS,
    PARAM_ACTIONS_FILENAME_NO_OF_MODEL_GROUPS_UNPLACED,
    PARAM_ACTIONS_FILENAME_NO_OF_DETAIL_GROUPS,
    PARAM_ACTIONS_FILENAME_NO_OF_DETAIL_GROUPS_UNPLACED,
    PARAM_ACTIONS_FILENAME_NO_OF_ROOMS,
    PARAM_ACTIONS_FILENAME_NO_OF_ROOMS_UNPLACED,
    PARAM_ACTIONS_FILENAME_NO_OF_ROOMS_UNENCLOSED,
    PARAM_ACTIONS_FILENAME_NO_OF_ROOMS_REDUNDANT,
    PARAM_ACTIONS_FILENAME_NO_OF_FILLED_REGIONS,
    PARAM_ACTIONS_FILENAME_DATE_LAST_UPDATED,
    PARAM_ACTIONS_FILENAME_NO_OF_ARROW_HEAD_STYLES,
    PARAM_ACTIONS_FILENAME_NO_OF_ARROW_HEAD_STYLES_UNUSED,
    PARAM_ACTIONS_FILENAME_NO_OF_TEXT_STYLES,
    PARAM_ACTIONS_FILENAME_NO_OF_TEXT_STYLES_UNUSED,
    PARAM_ACTIONS_FILENAME_NO_OF_DIMENSION_STYLES,
    PARAM_ACTIONS_FILENAME_NO_OF_DIMENSION_STYLES_UNUSED,
}



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A module with helper function around family types catalogue files.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Family types can be filtered ( removed from the export ) by providing a list of filters.
For available filters see the documentation in the module duHast.Utilities.compare alternatively inline filters can be provided as lambda functions.

 filters = [[lambda x, y: x.startswith(y), "A"], [lambda x, y: x.endswith(y), "B"]]

The filters are applied to the family type names ( x in the above examples ). If the type name passes all filters it will be added to the catalogue file.


"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


import os
import csv

from duHast.Revit.Family.family_types_get_data_from_xml import get_type_data_via_XML_from_family_file
from duHast.Revit.Family.family_parameter_utils import  get_family_type_parameters,  filter_parameters_by_formula_driven
from duHast.Revit.Family.Data.Objects.family_type_data_storage_manager import FamilyTypeDataStorageManager

from duHast.Utilities.Objects.result import Result
from duHast.Utilities.files_io import get_file_name_without_ext, get_directory_path_from_file_path,file_exist
from duHast.Utilities.files_csv import write_report_data_as_csv
from duHast.Utilities.Objects.file_encoding_bom import BOMValue


def type_name_passes_filters(fam_type, filters):

    """
    Check if the family type name passes the filters.

    :param fam_type: The family type to check.
    :type fam_type: FamilyTypeDataStorage
    :param filters: The filters to apply to the types to be added to the catalogue file. If None, no filters will be applied and all types will be added. If more then one filter is provided, the types must pass all filters
    :type filters: list(list(function(value1,value2), test_value))

    :return: True if the type name passes all filters, False otherwise.
    :rtype: bool
    """

    # check if filters are None
    if filters is None:
        return True
    
    # go over filters and check if the type passes all filters by type name
    passes_filter = True

    for filter_instance in filters:

        # do some sanity checks
        if not isinstance(filter_instance, list):
            return False

        if len(filter_instance) != 2:
            return False

        # get the filter function
        filter_func = filter_instance[0]
        filter_value = filter_instance[1]

        # check if the filter function is callable
        if not callable(filter_func):
            return False

        # check if the type name passes the filter
        passes_filter = passes_filter and filter_func(fam_type.family_type_name, filter_value)

    return passes_filter

   
def pass_fam_types_through_filters(fam_type_manager, filters):
    """
    Pass family types through filters.

    :param fam_type_manager: The family type manager to pass through the filters.
    :type fam_type_manager: FamilyTypeDataStorageManager
    :param filters: The filters to apply to the types to be added to the catalogue file. If None, no filters will be applied and all types will be added. If more then one filter is provided, the types must pass all filters
    :type filters: list(list(function(value1,value2), test_value))
    
    :return: The updated family type manager.
    :rtype: FamilyTypeDataStorageManager
    """

    return_value = Result()

    try:
        # loop over filters
        for filter_instance in filters:

            # do some sanity checks
            if not isinstance(filter_instance, list):
                return_value.update_sep(False, "Filter is not a list: {}".format(filter_instance))
                continue

            if len(filter_instance) != 2:
                return_value.update_sep(False, "Filter does not contain two elements: {}".format(filter_instance))
                continue

            # get the filter function
            filter_func = filter_instance[0]

            # check if the filter function is callable
            if not callable(filter_func):
                return_value.update_sep(False, "Filter function is not callable: {}".format(filter_func))
                continue
        
            # loop over family types and remove types not passing filters
            for fam_type_storage in fam_type_manager.family_type_data_storage:
                if (not type_name_passes_filters( fam_type_storage, filters)):
                    removed_type_flag = fam_type_manager.remove_family_type( fam_type_storage.family_type_name)
                    return_value.update_sep( removed_type_flag, "Removed family type: {} with status: {}".format(fam_type_storage.family_type_name, removed_type_flag))
                else:
                    return_value.append_message(True, "Family type: {} passed filters.".format(fam_type_storage.family_type_name))

            
            # return the updated family type manager
            return_value.result.append(fam_type_manager)

    except Exception as e:
        return_value.update_sep(False, "Failed to pass family types through filters: {}".format(str(e)))

    return return_value


def get_type_parameter_order(doc, parameter_order):
    """
    Get the type parameter order for the catalogue file.
    
    :param doc: The family document to extract the type data from.
    :type doc: rdb.Family
    
    :param parameter_order: The order of the parameters in the catalogue file. If None, the parameters will be ordered alphabetically.
    :type parameter_order: list(str)
    
    :return: The type parameter order for the catalogue file with the result class .result property
    :rtype: Result
    """

    return_value = Result()
    
    try:
        # get list of parameters to export ( remove any type driven parameters governed by a formula)
        # get type parameters
        type_parameters = get_family_type_parameters(doc)

        # check if the type parameters are valid
        if type_parameters is None or len(type_parameters) == 0:
            return_value.update_sep(False, "Failed to get the family type parameters.")
            return return_value
        
        # filter out formula driven parameters
        type_parameters_filtered = filter_parameters_by_formula_driven(type_parameters, False)

        # check if the filtered type parameters are valid
        if type_parameters_filtered is None or len(type_parameters_filtered) == 0:
            return_value.update_sep(False, "Filter formula driven parameters removed all parameters from the set.")
            return return_value
        
        # get the parameter order
        if parameter_order is None:
            return_value.result = sorted([param.Definition.Name for param in type_parameters_filtered])
        else:
            family_parameter_names_sorted =  sorted([param.Definition.Name for param in type_parameters_filtered])
            for para_name in parameter_order:
                if para_name not in family_parameter_names_sorted:
                    return_value.append_message("Parameter {} not found in type parameters.".format(para_name))
                else:
                    return_value.result.append(para_name)
                    # remove that parameter from the list
                    family_parameter_names_sorted.remove(para_name)
            # check if any parameters are left
            if len(family_parameter_names_sorted) > 0:
                # add the rest of the parameters
                return_value.result = return_value.result + family_parameter_names_sorted

    except Exception as e:
        return_value.update_sep(False, "Failed to setup parameter order: {}".format(str(e)))

    return return_value


def write_catalogue_file_to_csv(catalogue_file_data, family_file_path, header, override_existing = False):
    """
    Write the catalogue file to disk.

    :param catalogue_file_data: The catalogue file data to write to disk.
    :type catalogue_file_data: list(list(str))
    :param family_file_path: The path to the family file.
    :type family_file_path: str
    :param override_existing: If True, the existing catalogue file will be overwritten. If False, no catalogue file will be exported.
    :type override_existing: bool

    :return: The result of the write operation.
    :rtype: Result
    """

    return_value = Result()

    try:
        # build the catalogue file path
        catalogue_file_directory = get_directory_path_from_file_path( family_file_path)
        catalogue_file_path_name = "{}.txt".format(get_file_name_without_ext( family_file_path))
        catalogue_file_full_path = os.path.join(catalogue_file_directory, catalogue_file_path_name)

        # check if the file already exists
        if file_exist(catalogue_file_full_path) and not override_existing:
            return_value.update_sep(False, "Catalogue file already exists and override_existing is False.")
            return return_value
        
        # write the catalogue file to disk
        write_result = write_report_data_as_csv(
            file_name=catalogue_file_full_path,
            header=header, 
            data= catalogue_file_data,
            encoding="utf-16-le",
            bom=BOMValue.UTF_16_LITTLE_ENDIAN, 
            quoting=csv.QUOTE_MINIMAL
        )

        return_value.update(write_result)
        
    except Exception as e:
        return_value.update_sep(False, "An exception occurred when trying to write catalogue file to disk: {}".format(e))

    return return_value


def export_catalogue_file(doc, file_path = None, filters = None, parameter_order = None, override_existing = False):
    """
    Export the family types catalogue file.

    :param doc: The family document to extract the type data from.
    :type doc: rdb.Family
    :param file_path: The path to save the catalogue file. If None, the file will be saved in the same location as the family with the same name as the family.
    :type file_path: str
    :param filters: 
        
            The filters to apply to the types to be added to the catalogue file. If None, no filters will be applied and all types will be added. If more then one filter is provided, the types must pass all filters to be added.
            Filter format:

                [[func(family_type_name, filter value), "my check value"],...]
            
            Example:
                filters = [[lambda x, y: x.startswith(y), "A"], [lambda x, y: x.endswith(y), "B"]]

    :type filters: list(function(value1,value2))
    :param parameter_order: The order of the parameters in the catalogue file. If None, the parameters will be ordered alphabetically.
    :param override_existing: If True, the existing catalogue file will be overwritten. If False, no catalogue file will be exported.
    """

    return_value = Result()

    try:

        # check if a family document...
        if not doc.IsFamilyDocument:
            return_value.update_sep(False, "The document is not a family document.")
            return return_value

        # get the family name
        family_name = doc.Title

        # remove the file extension
        if family_name.lower().endswith(".rfa"):
            family_name = family_name[:-4]

        # get the family path
        family_path = doc.PathName

        # check if the file path is provided
        if file_path is None:
            file_path= family_path

        # get the family type data
        family_type_data_result = get_type_data_via_XML_from_family_file(doc.Application, family_name, family_path)

        # check if the family type data was successfully extracted
        if not family_type_data_result.status:
            return_value.update_sep(False, "Failed to get the family type data: {}".format(family_type_data_result.message))
            return return_value
        
        if family_type_data_result.result is None or len(family_type_data_result.result) == 0:
            return_value.update_sep(False, "No family type data was extracted.")
            return return_value
        
        # get the family type data
        fam_type_manager = family_type_data_result.result[0]

        if not isinstance(fam_type_manager, FamilyTypeDataStorageManager):
            return_value.update_sep(False, "Failed to get the family type manager. Got {} instead.".format(type(fam_type_manager)))
            return return_value
        
        # check if family has any types
        if not fam_type_manager.family_has_types:
            return_value.update_sep(False, "No family types found in family.")
            return return_value

        # go over filters and remove types that do not pass the filter
        if filters is not None:

            # attempt to pass family types through filters
            fam_type_manager_filter_result = pass_fam_types_through_filters(fam_type_manager, filters)

            # check if the family types passed through the filters
            if (not fam_type_manager_filter_result.status):
                # if not get out
                return_value.update_sep(False, "Failed to pass family types through filters: {}".format(fam_type_manager_filter_result.message))
                return return_value
        

        # get the parameter order for the catalogue file
        # this will also remove any type parameters that are formula driven
        type_parameter_order_result = get_type_parameter_order(doc, parameter_order)
    
        if not type_parameter_order_result.status:
            return_value.update_sep(False, "Failed to get the type parameter order: {}".format(type_parameter_order_result.message))
            return return_value
        
        # get the type parameter order
        type_parameter_order = type_parameter_order_result.result

        # export the catalogue file       
        catalogue_file_data = fam_type_manager.get_catalogue_file_data(type_parameter_order)

        # check if the catalogue file data is valid
        if catalogue_file_data is None or len(catalogue_file_data) == 0:
            return_value.update_sep(False, "Failed to get the catalogue file data.")
            return return_value

        # build the header
        catalogue_file_header = fam_type_manager.get_catalogue_file_header_row(type_parameter_order)

        # check if header is valid
        if catalogue_file_header is None or len(catalogue_file_header) == 0:
            return_value.update_sep(False, "Failed to get the catalogue file header.")
            return return_value

        #print("Catalogue file data: {}".format(catalogue_file_data))
        #print("Catalogue file header: {}".format(catalogue_file_header))
        #print("Catalogue file path: {}".format(family_path))
        #print("Override existing: {}".format(override_existing))

        # write the catalogue file to file
        write_catalogue_file_result = write_catalogue_file_to_csv(
            catalogue_file_data=catalogue_file_data, 
            header=catalogue_file_header,
            family_file_path= family_path, 
            override_existing=override_existing
        )

        # check if the write was successful
        if not write_catalogue_file_result.status:
            return_value.update_sep(False, "Failed to write the catalogue file to file: {}".format(write_catalogue_file_result.message))
            return return_value

        return_value.update_sep(True, "Catalogue file successfully exported to: {}".format(family_path))

    except Exception as e:
        return_value.update_sep(False, str(e))

    return return_value
    


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A timer class to measure code performance.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from System.Diagnostics import Stopwatch
from duHast.Utilities.Objects import base

#: time output formatting 2 digit padding
PAD_SINGLE_DIGIT_TO_TWO = "%02d"
#: time output formatting default 3 digit padding
PAD_SINGLE_DIGIT_TO_THREE = "%03d"


class TimerError(Exception):
    """
    A custom exception used to report errors in use of Timer class
    """


class Timer(base.Base):
    """
    A class that provides functionality to measure the elapsed time between starting and stopping the timer.

    Example Usage:
    ```python
    timer = Timer()  # Create a new instance of the Timer class
    timer.start()  # Start the timer
    # Perform some operations
    timer.stop()  # Stop the timer and get the elapsed time
    ```

    Methods:
    - __init__(): Initializes a new instance of the Timer class.
    - start(): Starts a new timer.
    - stop(): Stops the timer and returns the elapsed time.
    - is_running(): Checks whether the timer is running.
    """

    def __init__(self):
        """
        Initializes a new instance of the Timer class.
        """
        self._stop_watch = None

        super(Timer, self).__init__()

    def start(self):
        """
        Starts a new timer.

        :raises TimerError: When the timer is already running.
        """
        if self._stop_watch is not None:
            raise TimerError("Timer is running. Use .stop() to stop it")

        self._stop_watch = Stopwatch()
        self._stop_watch.Start()

    def stop(self):
        """
        Stops the timer and returns the elapsed time.

        :raises TimerError: When the timer is not running yet.
        :return: The elapsed time since the timer has started.
        :rtype: str
        """
        if self._stop_watch is None:
            raise TimerError("Timer is not running. Use .start() to start it")
        self._stop_watch.Stop()
        time_span = self._stop_watch.Elapsed
        self._stop_watch = None
        return (
            "Elapsed time: "
            + str(PAD_SINGLE_DIGIT_TO_TWO % time_span.Hours)
            + "h."
            + str(PAD_SINGLE_DIGIT_TO_TWO % time_span.Minutes)
            + "m."
            + str(PAD_SINGLE_DIGIT_TO_TWO % time_span.Seconds)
            + "s."
            + str(PAD_SINGLE_DIGIT_TO_THREE % time_span.Milliseconds)
            + "ms"
        )

    def is_running(self):
        """
        Checks whether the timer is running.

        :return: True if the timer is running, False otherwise.
        :rtype: bool
        """
        if self._stop_watch is None:
            return False
        else:
            return True



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit curve helper functions
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from System.Collections.Generic import List

import math

from Autodesk.Revit.DB import Arc, CurveLoop, ElementId, Line, ModelCurve, Transform, XYZ

from duHast.Utilities.compare import is_close
from duHast.Utilities.Objects import result as res


def is_curve_is_within_curve(curve_one, curve_two):
    """
    Checks whether one curve is completely within another curve.

    Args:
        curve_one: The first curve to compare (can be an arc or a line).
        curve_two: The second curve to compare (can be an arc or a line).

    Returns:
        The curve that is completely within the other curve, or None if the curves are overlapping.

    Example Usage:
        curve_one = ...
        curve_two = ...
        result = check_curve_is_within_curve(curve_one, curve_two)
        print(result)

        Output:
        - If the curves are identical, it will return curve_one.
        - If curve_two is completely within curve_one, it will return curve_two.
        - If curve_one is completely within curve_two, it will return curve_one.
        - If the curves are overlapping, it will return None.
    """

    curve_one_end_zero_point = curve_one.curve.GetEndPoint(0)
    curve_one_end_one_point = curve_one.curve.GetEndPoint(1)
    curve_two_end_zero_point = curve_two.curve.GetEndPoint(0)
    curve_two_end_one_point = curve_two.curve.GetEndPoint(1)

    # distance of close to zero indicates point is on line!
    distance_one_end_zero_to_two = round(
        curve_two.curve.Distance(curve_one_end_zero_point), 4
    )
    distance_one_end_one_to_two = round(
        curve_two.curve.Distance(curve_one_end_one_point), 4
    )
    distance_two_end_zero_to_one = round(
        curve_one.curve.Distance(curve_two_end_zero_point), 4
    )
    distance_two_end_one_to_one = round(
        curve_one.curve.Distance(curve_two_end_one_point), 4
    )

    # possible out comes:
    # curves are equal in length and location ( all 4 tests return true )
    # one curve is fully inside another ( 1 pair test result is false, the other pair is true )
    # curves are overlapping ( both pairs return false )

    if (
        is_close(distance_one_end_zero_to_two, 0.000, 0.000)
        and is_close(distance_one_end_one_to_two, 0.000, 0.000)
        and is_close(distance_two_end_zero_to_one, 0.000, 0.000)
        and is_close(distance_two_end_one_to_one, 0.000, 0.000)
    ):
        # curves are identical
        return curve_one
    elif is_close(distance_two_end_zero_to_one, 0.000, 0.000) and is_close(
        distance_two_end_one_to_one, 0.000, 0.000
    ):
        # curve two is completely within curve one
        return curve_two
    elif is_close(distance_one_end_zero_to_two, 0.000, 0.000) and is_close(
        distance_one_end_one_to_two, 0.000, 0.000
    ):
        # curve one is completely within curve two
        return curve_one
    else:
        # curves are overlapping
        return None


def calculate_lengthened_curve_geometry(curve_one, curve_two):
    """
    Two options when lines are overlapping:

    - extend the longer line to the end of the shorter line not overlapping and delete the shorter line or
    - shorten shorter line to move the endpoint currently overlapping to the end point of the longer line, therefore removing the overlap

    First option will shorten number of room separation lines in model. However when multiple lines overlap a single long line this
    code will delete a shorter overlapping line still required!

    returns two values: first one is the curve to change in length, second one is the curve to delete!
    """

    # check first if this is a complete overlap (couldn't fixed in step one since smaller line is in group...)
    # if so dont make any changes here...
    check_complete_overlap = is_curve_is_within_curve(curve_one, curve_two)
    if check_complete_overlap == None:
        # which curve is the longer curve?
        # assume default
        longer_curve = curve_two
        shorter_curve = curve_one
        # check default was correct?
        if curve_one.curve.ApproximateLength > curve_two.curve.ApproximateLength:
            longer_curve = curve_one
            shorter_curve = curve_two

        # find the endpoint of the shorter line which is not within the longer line
        shorter_curve_end_zero = shorter_curve.curve.GetEndPoint(0)
        shorter_curve_end_one = shorter_curve.curve.GetEndPoint(1)

        # measure distance to other curve. Due to precision issues these
        # values are rounded. A distance of 0 indicates point is on curve
        distance_shorter_curve_end_zero_to_longer_curve = round(
            longer_curve.curve.Distance(shorter_curve_end_zero), 4
        )
        distance_shorter_end_one_to_longer_curve = round(
            longer_curve.curve.Distance(shorter_curve_end_one), 4
        )

        # check which end of the shorter curve is outside the longer curve
        is_shorter_curve_end_zero_on_longer_curve = is_close(
            distance_shorter_curve_end_zero_to_longer_curve, 0.000, 0.000
        )
        is_shorter_curve_end_one_on_longer_curve = is_close(
            distance_shorter_end_one_to_longer_curve, 0.000, 0.000
        )

        # check which end of the longer curve is outside the shorter curve ( that end will stay unchanged )
        # find the endpoint of the shorter line which is not within the longer line
        longer_curve_end_zero = longer_curve.curve.GetEndPoint(0)
        longer_curve_end_one = longer_curve.curve.GetEndPoint(1)

        # measure distance to other curve. Due to precision issues these
        # values are rounded. A distance of 0 indicates point is on curve
        distance_longer_curve_end_zero_to_shorter_curve = round(
            shorter_curve.curve.Distance(longer_curve_end_zero), 4
        )
        distance_longer_curve_end_one_to_shorter_curve = round(
            shorter_curve.curve.Distance(longer_curve_end_one), 4
        )

        # check which end of the longer curve is outside the shorter curve
        is_longer_curve_end_zero_on_shorter_curve = is_close(
            distance_longer_curve_end_zero_to_shorter_curve, 0.000, 0.000
        )
        is_longer_curve_end_one_on_shorter_curve = is_close(
            distance_longer_curve_end_one_to_shorter_curve, 0.000, 0.000
        )

        # determine the start and end point of the longer curve
        point_zero = None
        point_one = None
        # check which point of the shorter curve is not on the longer curve ( new end point )
        if is_shorter_curve_end_zero_on_longer_curve:
            # one end of shorter curve is outside the longer curve
            point_one = shorter_curve.curve.GetEndPoint(1)
        elif is_shorter_curve_end_one_on_longer_curve:
            # zero end of shorter curve is outside the longer curve
            point_one = shorter_curve.curve.GetEndPoint(0)
        else:
            raise ValueError(
                "Neither end point of shorter curve {} is on longer curve {}. That is impossible since that would mean curves are not overlapping...at all.\nDistance shorter curve end point zero to longer curve: {}\nDistance shorter curve end point one to longer curve: {}".format(
                    shorter_curve.id,
                    longer_curve.id,
                    distance_shorter_curve_end_zero_to_longer_curve,
                    distance_shorter_end_one_to_longer_curve,
                )
            )

        # check which point of the longer curve is not on the shorter curve ( new start point )
        if is_longer_curve_end_zero_on_shorter_curve:
            # point at end 1 is not one shorter curve
            point_zero = longer_curve.curve.GetEndPoint(1)
        elif is_longer_curve_end_one_on_shorter_curve:
            # point at end 0 is not one shorter curve
            point_zero = longer_curve.curve.GetEndPoint(0)
        else:
            raise ValueError(
                "Neither end point of longer curve {} is on shorter curve {}.\n This indicates the longer curve is completely overlapping the shorter curve. \nDistance longer curve end point zero to shorter curve: {}\nDistance longer curve end point one to shorter curve: {}\nDistance shorter curve end point zero to longer curve: {}\nDistance shorter curve end point one to longer curve: {}".format(
                    longer_curve.id,
                    shorter_curve.id,
                    distance_longer_curve_end_zero_to_shorter_curve,
                    distance_longer_curve_end_one_to_shorter_curve,
                    distance_shorter_curve_end_zero_to_longer_curve,
                    distance_shorter_end_one_to_longer_curve,
                )
            )

        new_point_zero = None
        new_point_one = None
        # trying to avoid to flip start at end point of the line, since this may cause the rooms to moveI(!?)
        if longer_curve.curve.GetEndPoint(0).IsAlmostEqualTo(point_zero):
            # assign the same end point for 0 index
            new_point_zero = longer_curve.curve.GetEndPoint(0)
            # assign a new end point for 1 index
            new_point_one = point_one
        elif longer_curve.curve.GetEndPoint(0).IsAlmostEqualTo(point_one):
            # assign the same end point for 0 index
            new_point_zero = longer_curve.curve.GetEndPoint(0)
            # assign a new end point for 1 index
            new_point_one = point_zero
        elif longer_curve.curve.GetEndPoint(1).IsAlmostEqualTo(point_zero):
            # assign a new end point for 0 index
            new_point_zero = point_one
            # assign the same end point for 1 index
            new_point_one = longer_curve.curve.GetEndPoint(1)
        else:
            # assign a new end point for 0 index
            new_point_zero = point_zero
            # assign the same end point for 1 index
            new_point_one = longer_curve.curve.GetEndPoint(1)

        # check for to short lines:
        # Aborted with exception: Curve length is too small for Revit's tolerance (as identified by Application.ShortCurveTolerance).

        # create a new curve depending on whether this is an arc or a line
        if type(longer_curve.curve) == Line:
            try:
                longer_curve.new_curve = Line.CreateBound(new_point_zero, new_point_one)
            except Exception as e:
                print(
                    "new line curve: {} points: 0: {} 1: {} with exception: {}".format(
                        longer_curve.id, new_point_zero, new_point_one, e
                    )
                )
        elif type(longer_curve.curve) == Arc:
            try:
                longer_curve.new_curve = Arc.Create(
                    new_point_zero, new_point_one, longer_curve.curve.Center
                )
            except Exception as e:
                print(
                    "new arc curve: {} points: 0: {} 1: {} C: {} with exception: {}".format(
                        longer_curve.id,
                        new_point_zero,
                        new_point_one,
                        longer_curve.curve.Center,
                        e,
                    )
                )
        else:
            raise ValueError(
                "Curve type {} is not supported.".format(type(longer_curve.curve))
            )

        return longer_curve, shorter_curve
    else:
        return None, None


def calculate_shortened_curve_geometry(curve_one, curve_two):
    """
    Two options when lines are overlapping:

    - extend the longer line to the end of the shorter line not overlapping and delete the shorter line or
    - shorten shorter line to move the endpoint currently overlapping to the end point of the longer line, therefore removing the overlap

    This is an attempt on option 2

    returns the curve of which to change the length!
    """
    # which curve is the longer curve?
    # assume default
    longer_curve = curve_two
    shorter_curve = curve_one
    # check default was correct?
    if curve_one.curve.ApproximateLength > curve_two.curve.ApproximateLength:
        longer_curve = curve_one
        shorter_curve = curve_two

    # find the endpoint of the shorter line which is not within the longer line
    shorter_curve_end_zero = shorter_curve.curve.GetEndPoint(0)
    shorter_curve_end_one = shorter_curve.curve.GetEndPoint(1)

    # measure distance to other curve. Due to precision issues these
    # values are rounded. A distance of 0 indicates point is on curve
    distance_shorter_curve_end_zero_to_longer_curve = round(
        longer_curve.curve.Distance(shorter_curve_end_zero), 4
    )
    distance_shorter_end_one_to_longer_curve = round(
        longer_curve.curve.Distance(shorter_curve_end_one), 4
    )

    # check which end of the shorter curve is outside the longer curve
    is_shorter_curve_end_zero_on_longer_curve = is_close(
        distance_shorter_curve_end_zero_to_longer_curve, 0.000, 0.000
    )
    is_shorter_curve_end_one_on_longer_curve = is_close(
        distance_shorter_end_one_to_longer_curve, 0.000, 0.000
    )

    # check which end of the longer curve is outside the shorter curve ( that end will stay unchanged )
    # find the endpoint of the shorter line which is not within the longer line
    longer_curve_end_zero = longer_curve.curve.GetEndPoint(0)
    longer_curve_end_one = longer_curve.curve.GetEndPoint(1)

    # measure distance to other curve. Due to precision issues these
    # values are rounded. A distance of 0 indicates point is on curve
    distance_longer_curve_end_zero_to_shorter_curve = round(
        shorter_curve.curve.Distance(longer_curve_end_zero), 4
    )
    distance_longer_curve_end_one_to_shorter_curve = round(
        shorter_curve.curve.Distance(longer_curve_end_one), 4
    )

    # check which end of the longer curve is outside the shorter curve
    is_longer_curve_end_zero_on_shorter_curve = is_close(
        distance_longer_curve_end_zero_to_shorter_curve, 0.000, 0.000
    )
    is_longer_curve_end_one_on_shorter_curve = is_close(
        distance_longer_curve_end_one_to_shorter_curve, 0.000, 0.000
    )

    # determine the start and end point of the shorter curve
    point_zero = None
    point_one = None
    # check which point of the longer curve is on the shorter curve ( new end point )
    if is_longer_curve_end_zero_on_shorter_curve:
        # one end of longer curve is on the shorter curve
        point_one = longer_curve.curve.GetEndPoint(0)
    elif is_longer_curve_end_one_on_shorter_curve:
        # zero end of longer curve is on the shorter curve
        point_one = longer_curve.curve.GetEndPoint(1)
    else:
        raise ValueError(
            "Neither end point of {} is on {}".format(shorter_curve.id, longer_curve.id)
        )

    # check which point of the shorter curve is not on the longer curve ( start point to keep)
    if is_shorter_curve_end_zero_on_longer_curve:
        # point at end 1 is not one longer curve
        point_zero = shorter_curve.curve.GetEndPoint(1)
    elif is_shorter_curve_end_one_on_longer_curve:
        # point at end 0 is not one longer curve
        point_zero = shorter_curve.curve.GetEndPoint(0)
    else:
        raise ValueError(
            "Neither end point of {} is on {}".format(longer_curve.id, shorter_curve.id)
        )

    new_point_zero = None
    new_point_one = None
    # trying to avoid to flip start at end point of the line, since this may cause the rooms to moveI(!?)
    if shorter_curve.curve.GetEndPoint(0).IsAlmostEqualTo(point_zero):
        # assign the same end point for 0 index
        new_point_zero = shorter_curve.curve.GetEndPoint(0)
        # assign a new end point for 1 index
        new_point_one = point_one
    elif shorter_curve.curve.GetEndPoint(0).IsAlmostEqualTo(point_one):
        # assign the same end point for 0 index
        new_point_zero = shorter_curve.curve.GetEndPoint(0)
        # assign a new end point for 1 index
        new_point_one = point_zero
    elif shorter_curve.curve.GetEndPoint(1).IsAlmostEqualTo(point_zero):
        # assign a new end point for 0 index
        new_point_zero = point_one
        # assign the same end point for 1 index
        new_point_one = shorter_curve.curve.GetEndPoint(1)
    else:
        # assign a new end point for 0 index
        new_point_zero = point_zero
        # assign the same end point for 1 index
        new_point_one = shorter_curve.curve.GetEndPoint(1)

    # create a new curve depending on whether this is an arc or a line
    if type(shorter_curve.curve) == Line:
        shorter_curve.new_curve = Line.CreateBound(new_point_zero, new_point_one)
    elif type(shorter_curve.curve) == Arc:
        shorter_curve.new_curve = Arc.Create(
            new_point_zero, new_point_one, shorter_curve.curve.Centre
        )
    else:
        raise ValueError(
            "Curve type {} is not supported.".format(type(longer_curve.curve))
        )

    return shorter_curve


def modify_model_line_action(existing_curve, new_curve, override_joins=True):
    """
    Modify the geometry of an existing curve by setting it to a new curve.

    Args:
        existing_curve (object): The existing curve that needs to be modified.
        new_curve (object): The new curve that will replace the existing curve.
        override_joins (bool, optional): Flag to indicate whether to override the joins between adjoining model lines. Defaults to True.

    Returns:
        object: An instance of the Result class that contains information about the success or failure of the geometry modification.
    """
    return_value = res.Result()
    try:
        # make sure to set this to True, otherwise adjoining model lines will move with end points
        existing_curve.SetGeometryCurve(new_curve, override_joins)
        return_value.update_sep(
            True,
            "Successfully changed geometry of element: {}".format(existing_curve.Id),
        )
    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to change geometry of element {} with exception: {}".format(
                existing_curve.Id, e
            ),
        )
    return return_value


def translate_curves_in_elevation(
    doc, original_curves, source_elevation, target_elevation
):
    return_value = res.Result()
    try:
        new_curves = doc.Application.Create.NewCurveArray()
        # work out Z translation vector
        z = target_elevation - source_elevation
        return_value.append_message(
            "new level RL: {} old level RL: {} translation z: {}".format(
                target_elevation, source_elevation, z
            )
        )
        # vector of translation
        vector_translation = XYZ(0, 0, z)
        # transformation
        tf = Transform.CreateTranslation(vector_translation)
        for sgs in original_curves:
            new_curves.Append(sgs.CreateTransformed(tf))
        return_value.append_message("Created {} mew curve(s)".format(new_curves.Size))
        return_value.result = [new_curves]
    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to translate curves from elevation {} to elevation: {} with exception: {}".format(
                source_elevation, target_elevation, e
            ),
        )
    return return_value


def get_perpendicular_line_through_point(original_line, rotation_axis, rotation_point):
    """
    Returns a line rotated by 90 degree around the rotation point.

    :param original_line: The original line to be rotated
    :type original_line:  Autodesk.Revit.DB.Line
    :param rotation_point: The point to rotate around.
    :type rotation_point:  Autodesk.Revit.DB.XYZ
    :param rotation_axis: The axis around to rotate the line
    :type rotation_axis:  Autodesk.Revit.DB.XYZ

    :return: The rotated line.
    :rtype: Autodesk.Revit.DB.Line
    """

    # get a line through origin at 90 degree
    trans_form = Transform.CreateRotationAtPoint(
        rotation_axis, math.pi / 2.0, rotation_point
    )
    transformed_line = original_line.CreateTransformed(trans_form)
    return transformed_line


def get_curve_level(doc, curve):
    """
    Sorts the area lines based on their associated level names and returns a dictionary where the keys are the level names and the values are lists of area lines.

    Args:
        doc (Document): The current model document.
        area_lines (list): A list of area lines.

    Returns:
        dict: A dictionary where the keys are the level names and the values are lists of area lines.
    """

    if isinstance(curve, ModelCurve)==False:
        raise TypeError("Expect curve to be a ModelCurve instance. Got {} instead".format(type(curve)))
    
    level_id = curve.LevelId
    if(level_id==ElementId.InvalidElementId):
        return None
    level = doc.GetElement(level_id)
    return level


def are_lines_parallel(line_one, line_two):
    """
    Check if two lines are parallel.

    Args:
        line_one (object): The first line to compare.
        line_two (object): The second line to compare.

    Returns:
        bool: True if the lines are parallel, False otherwise.
    """
    # Get normalized directions for both lines
    dir_one = line_one.Direction.Normalize()
    dir_two = line_two.Direction.Normalize()

    # Check if the directions are the same or opposite
    return dir_one.IsAlmostEqualTo(dir_two) or dir_one.IsAlmostEqualTo(dir_two.Negate())

def are_lines_perpendicular(line_one, line_two):
    """
    Check if two lines are perpendicular.

    Args:
        line_one (object): The first line to compare.
        line_two (object): The second line to compare.

    Returns:
        bool: True if the lines are perpendicular, False otherwise.
    """

    # check if the lines are perpendicular:
    # calculate a perpendicular vector to the first line
    # and check if it is parallel to the second line
    # thats done rather than using the dot product of the two lines
    # to stay within the means of Revit API calculation precision and avoid floating point errors
    perpendicular_vector_to_one = XYZ(-line_one.Direction.Y, line_one.Direction.X, 0)
    
    # normalize the direction of the second line
    dir_two = line_two.Direction.Normalize()

    # check if the perpendicular vector is parallel to the second line
    return perpendicular_vector_to_one.IsAlmostEqualTo(dir_two) or perpendicular_vector_to_one.IsAlmostEqualTo(dir_two.Negate())


def offset_curve(curve, normal_vector, offset_distance):
    """
    Offsets a curve in the direction of the normal vector by a specified distance.

    :param curve: The curve to be offset.
    :type curve: Autodesk.Revit.DB.Curve
    :param normal_vector: The normal vector indicating the offset direction.
    :type normal_vector: Autodesk.Revit.DB.XYZ
    :param offset_distance: The distance to offset the curve.
    :type offset_distance: float
    
    :return: The offset curve.
    :rtype: Autodesk.Revit.DB.Curve
    """
    
    # Make sure the normal vector is pointing towards offset direction
    if normal_vector.DotProduct(XYZ(0, 0, 1)) < 0:
        normal_vector = normal_vector.Negate()
    
    # Create offset vector
    offset_vector = XYZ(normal_vector.X, normal_vector.Y, 0).Normalize().Multiply(offset_distance)
    
    # Transform curve
    transform = Transform.CreateTranslation(offset_vector)
    return curve.CreateTransformed(transform)


def splice_adjacent_lines(curve_1, curve_2):
    """
    Splices two adjacent lines if they share an endpoint.

    :param curve_1: The first line to splice.
    :type curve_1: Autodesk.Revit.DB.Line
    :param curve_2: The second line to splice.
    :type curve_2: Autodesk.Revit.DB.Line
    :return: A new line that combines the two lines if they are adjacent, otherwise None.
    :rtype: Autodesk.Revit.DB.Line
    """

    # make sure we got lines
    if not isinstance(curve_1, Line) or not isinstance(curve_2, Line):
        # if not, return None
        raise TypeError("Both curves must be of type Line. curve_1 {} curve_2 {}".format(type(curve_1), type(curve_2)))
    
    combined_curve = None

    # check which curves points are almost equal to each other
    if curve_1.GetEndPoint(1).IsAlmostEqualTo(curve_2.GetEndPoint(0)):
        combined_curve = Line.CreateBound(curve_1.GetEndPoint(0), curve_2.GetEndPoint(1))
    elif curve_1.GetEndPoint(1).IsAlmostEqualTo(curve_2.GetEndPoint(1)):
        combined_curve = Line.CreateBound(curve_1.GetEndPoint(0), curve_2.GetEndPoint(0))
    elif curve_1.GetEndPoint(0).IsAlmostEqualTo(curve_2.GetEndPoint(1)):
        combined_curve = Line.CreateBound(curve_1.GetEndPoint(1), curve_2.GetEndPoint(0))
    elif curve_1.GetEndPoint(0).IsAlmostEqualTo(curve_2.GetEndPoint(0)):
        combined_curve = Line.CreateBound(curve_1.GetEndPoint(1), curve_2.GetEndPoint(1))
    
    
    return combined_curve


# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from duHast.Utilities.Objects.base import Base

from duHast.Revit.ColourFillSchemes.Reporting.colour_fill_scheme_defaults import (
    COLOUR_FILL_SCHEME_CSV_PROPERTY_INDEX,
    PROPERTY_FILL_SCHEME_NAME,
    PROPERTY_AREA_SCHEME_NAME,
    PROPERTY_PARAMETER_VALUE,
    PROPERTY_STORAGE_TYPE,
    PROPERTY_FILL_PATTERN_ID,
    PROPERTY_IS_IN_USE,
    PROPERTY_IS_VISIBLE,
    PROPERTY_COLOUR_RED,
    PROPERTY_COLOUR_GREEN,
    PROPERTY_COLOUR_BLUE,
)

class ColourFillStorage(Base):
    """
    A class to store colour fill information for objects.
    """

    def __init__(self):

        # ini super class to allow multi inheritance in children!
        super(ColourFillStorage, self).__init__()

        self.fill_scheme_name = ""
        self.area_scheme_name = ""
        self.parameter_value = ""
        self.fill_pattern_id = -1
        self.is_in_use=True
        self.is_visible=True
        self.storage_type=0
        self.colour_red=0
        self.colour_green=0
        self.colour_blue=0

    
    def import_from_data_row(self, data_row):
        """
        Imports data from a data row into the ColourFillStorage object.

        :param data_row: The data row to import from.
        :type data_row: list
        :raises TypeError: If data_row is not a list.
        :raises ValueError: If data_row does not have the correct number of elements.
        :raises TypeError: If data_row is not a list of strings.

        :return: None
        :rtype: None
        """

        # type checking:
        if not isinstance(data_row, list):
            raise TypeError("data_row must be of type list. Got {}".format(type(data_row)))

        if len(data_row) != len(COLOUR_FILL_SCHEME_CSV_PROPERTY_INDEX):
            raise ValueError("data_row must have {} elements. Got {}".format(len(COLOUR_FILL_SCHEME_CSV_PROPERTY_INDEX), len(data_row)))
        
        # check if the data_row is a list of strings
        if not all(isinstance(item, str) for item in data_row):
            raise TypeError("data_row must be a list of strings. Got {}".format(type(data_row)))
        
        # populate the properties from the data_row
        self.fill_scheme_name =data_row[COLOUR_FILL_SCHEME_CSV_PROPERTY_INDEX[PROPERTY_FILL_SCHEME_NAME]]
        self.area_scheme_name =data_row[COLOUR_FILL_SCHEME_CSV_PROPERTY_INDEX[PROPERTY_AREA_SCHEME_NAME]]
        self.parameter_value =data_row[COLOUR_FILL_SCHEME_CSV_PROPERTY_INDEX[PROPERTY_PARAMETER_VALUE]]
        self.storage_type = int(data_row[COLOUR_FILL_SCHEME_CSV_PROPERTY_INDEX[PROPERTY_STORAGE_TYPE]])
        self.fill_pattern_id = int(data_row[COLOUR_FILL_SCHEME_CSV_PROPERTY_INDEX[PROPERTY_FILL_PATTERN_ID]])
        self.is_in_use = bool(data_row[COLOUR_FILL_SCHEME_CSV_PROPERTY_INDEX[PROPERTY_IS_IN_USE]])
        self.is_visible = bool(data_row[COLOUR_FILL_SCHEME_CSV_PROPERTY_INDEX[PROPERTY_IS_VISIBLE]])
        self.colour_red = int(data_row[COLOUR_FILL_SCHEME_CSV_PROPERTY_INDEX[PROPERTY_COLOUR_RED]])
        self.colour_green = int(data_row[COLOUR_FILL_SCHEME_CSV_PROPERTY_INDEX[PROPERTY_COLOUR_GREEN]])
        self.colour_blue = int(data_row[COLOUR_FILL_SCHEME_CSV_PROPERTY_INDEX[PROPERTY_COLOUR_BLUE]])


    def get_report_headers(self):
        """
        Returns the headers for the report.
        
        :return: A list of headers.
        :rtype: list
        """
        return [
            "Fill Scheme Name",
            "Area Scheme Name", 
            "Parameter Value",
            "Parameter Storage Type",
            "Fill Pattern ID", 
            "Is In Use", 
            "Is Visible", 
            "Colour Red", 
            "Colour Green", 
            "Colour Blue"
        ]
    
    def get_report_data(self):
        """
        Returns the data for the report.
        
        :return: A list of data.
        :rtype: list
        """

        # get the value of the entry as a string depending on the storage type
        entry_value = "None"

        if self.storage_type == 0:
            entry_value = "None"
        elif self.storage_type == 1:
            entry_value = str(self.parameter_value)
        elif self.storage_type == 2:
            entry_value = str(self.parameter_value)
        elif self.storage_type == 3:
            entry_value = self.parameter_value
        elif self.storage_type == 4:
            entry_value = str(self.parameter_value.IntegerValue)


        return [
            self.fill_scheme_name,
            self.area_scheme_name,
            entry_value,
            str(self.storage_type),
            str(self.fill_pattern_id),
            str(self.is_in_use),
            str(self.is_visible),
            str(self.colour_red),
            str(self.colour_green),
            str(self.colour_blue)
        ]


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to purging Revit detail items. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from duHast.Revit.Family import purge_unused_family_types as rFamPurge
from duHast.Revit.Common import purge_utils as rPurgeUtils
from duHast.Revit.DetailItems import detail_items as rDetail
from duHast.Revit.DetailItems.Utility import (
    detail_items_type_sorting as rDetailTypeSort,
)

# -------------------------------- repeating detail types -------------------------------------------------------


def get_used_repeating_detail_type_ids(doc):
    """
    Gets all used repeating detail type ids in the model.
    Used: at least one instance of this type is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of element ids representing repeating detail types.
    :rtype: list Autodesk.Revit.DB.ElementIds
    """

    ids = rPurgeUtils.get_used_unused_type_ids(
        doc, rDetail.get_all_repeating_detail_type_ids_available, 1, 1
    )
    return ids


def get_unused_repeating_detail_type_ids(doc):
    """
    Gets all unused repeating detail type ids in the model.
    Unused: not one instance of this type is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list of element ids representing repeating detail types.
    :rtype: list Autodesk.Revit.DB.ElementIds
    """

    ids = rPurgeUtils.get_used_unused_type_ids(
        doc, rDetail.get_all_repeating_detail_type_ids_available, 0, 1
    )
    return ids


def get_unused_repeating_detail_type_ids_for_purge(doc):
    """
    Gets type ids off all unused repeating detail types in model.
    This method can be used to safely delete unused repeating detail types. In the case that no basic\
        wall instance using any of the types is placed, this will return all but one type id since\
        Revit requires at least one repeating detail type definition to be in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing all repeating detail types not in use.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = rPurgeUtils.get_used_unused_type_ids(
        doc, rDetail.get_all_repeating_detail_type_ids_available, 0, 1
    )
    all_ids = rDetail.get_all_repeating_detail_type_ids_available(doc)
    # need to keep at least one
    if len(all_ids) == len(ids):
        ids.pop(0)
    return ids


# -------------------------------- Detail families -------------------------------------------------------


def get_all_used_detail_symbol_ids(doc):
    """
    Gets all used detail symbol type ids in model.
    Used: at least one instance of this type is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of family symbol (type) ids.
    :rtype: list Autodesk.Revit.DB.ElementIds
    """

    ids = []
    dic = rDetailTypeSort.build_detail_type_ids_dictionary(
        rDetail.get_all_detail_types_by_category(doc)
    )
    if rDetail.ELEMENT_TYPE in dic:
        ids_unfiltered = dic[rDetail.FAMILY_SYMBOL]
        # check if used in repeating details
        ids_repeat_det = rDetail.get_all_repeating_detail_type_ids_available(doc)
        # print('ids used in repeating details ' + str(len(idsRepeatDet)))
        # get detail types used in repeating details only
        ids_of_details_used_repeat_details = (
            rDetail.get_detail_symbols_used_in_repeating_details(doc, ids_repeat_det)
        )
        # get detail types used in model
        ids_used_in_model = rPurgeUtils.get_used_unused_type_ids(
            doc, rDetail.get_all_detail_symbol_ids_available, 1
        )
        print("ids used in model: {} ".format(len(ids_used_in_model)))
        # built overall ids list
        for id in ids_of_details_used_repeat_details:
            if id not in ids:
                ids.append(id)
        for id in ids_used_in_model:
            if id not in ids:
                ids.append(id)
        return ids
    else:
        return []


def get_all_unused_detail_symbol_ids(doc):
    """
    Gets all unused detail symbol type ids in model.
    Unused: Not one instance of this type is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of family symbol (type) ids.
    :rtype: list Autodesk.Revit.DB.ElementIds
    """

    ids = []
    all_available_ids = rDetail.get_all_detail_symbol_ids_available(doc)
    all_used_ids = get_all_used_detail_symbol_ids(doc)
    for id in all_available_ids:
        if id not in all_used_ids:
            ids.append(id)
    return ids


def get_all_unused_detail_symbol_ids_for_purge(doc):
    """
    Gets type ids off all unused detail symbols (types) in model.
    This method can be used to safely delete all unused detail symbols (types) and families.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing all unused detail symbols and families not in use.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = rFamPurge.get_unused_in_place_ids_for_purge(
        doc, get_all_unused_detail_symbol_ids
    )
    return ids


def get_used_filled_region_type_ids(doc):
    """
    Gets all used filled region type ids in model.
    Used: at least one instance of this type is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of filled region type ids.
    :rtype: list Autodesk.Revit.DB.ElementIds
    """

    ids = []
    ids_all = rDetail.get_all_filled_region_type_ids_available(doc)
    for id in ids_all:
        el = doc.GetElement(id)
        dic = rDetailTypeSort.build_dependent_elements_dictionary(
            doc, el.GetDependentElements(None)
        )
        if "Autodesk.Revit.DB.FilledRegion" in dic:
            ids.append(id)
    return ids


def get_unused_filled_region_type_ids(doc):
    """'
    Gets all unused filled region type ids in model.
    Unused: Not one instance of this type is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of filled region type ids.
    :rtype: list Autodesk.Revit.DB.ElementIds
    """

    ids = []
    ids_all = rDetail.get_all_filled_region_type_ids_available(doc)
    for id in ids_all:
        el = doc.GetElement(id)
        dic = rDetailTypeSort.build_dependent_elements_dictionary(
            doc, el.GetDependentElements(None)
        )
        if "Autodesk.Revit.DB.FilledRegion" in dic == False:
            ids.append(id)
    return ids


def get_unused_filled_region_type_ids_for_purge(doc):
    """
    Gets ids off all unused filled region types in model.
    This method can be used to safely delete all unused filled region types in model. In the case that no filled\
        region instance using any of the types is placed, this will return all but one type id since\
        Revit requires at least one filled region type definition to be in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing all unused filled region types not in use.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = get_unused_filled_region_type_ids(doc)
    ids_all = rDetail.get_all_filled_region_type_ids_available(doc)
    # need to keep at least one
    if len(ids_all) == len(ids):
        ids.pop(0)
    return ids



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of functions around Extensible storage in Revit.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects.result import Result

from duHast.Revit.Common.adesk_info import VENDOR_ID
from duHast.Revit.ExtensibleSchemas.data_storage_create import create_project_data_storage
from duHast.Revit.ExtensibleSchemas.data_storage import find_data_storage
from duHast.Revit.ExtensibleSchemas.extensible_schemas import does_schema_exist, get_schema

from Autodesk.Revit.DB.ExtensibleStorage import AccessLevel, Schema, SchemaBuilder
from System import Guid


def create_schema(
    schema_name,
    schema_documentation,
    string_guid,
    vendor_id = VENDOR_ID,
    access_level_read=AccessLevel.Public,
    access_level_write=AccessLevel.Public,
    field_builder=None,
):
    """
    Basic schema creation sample (excludes any fields).
    Creates a scheme or returns the already existing scheme with the same guid.

    :param schema_name: The name of the schema
    :type schema_name: str
    :param schema_documentation: A short description of the schema
    :type schema_documentation: str
    :param string_guid: A guid. (unique identifier of this schema)
    :type string_guid: str

    :param access_level_read: Access level to schema for read operations. Default is Public.
    :type access_level_read: Autodesk.Revit.DB.ExtensibleStorage.AccessLevel

    :access_level_write: Access level to schema for write operations. Default is Public
    :access_level_write: Autodesk.Revit.DB.ExtensibleStorage.AccessLevel

    :return: The newly created schema, or existing schema with same guid.
    :rtype: Autodesk.Revit.DB.ExtensibleStorage.Schema
    """

    # check if the schema name contains any spaces or not
    if " " in schema_name:
        raise ValueError("Schema name cannot contain spaces.")
    
    # Look for schema in memory
    schema = Schema.Lookup(Guid(string_guid))
    # Check if schema exists in the memory or not
    if schema == None:
        # Create a schema builder
        schema_guid = Guid(string_guid)
        builder = SchemaBuilder(schema_guid)
        # Set read and write access levels
        builder.SetReadAccessLevel(access_level_read)
        builder.SetWriteAccessLevel(access_level_write)
        # Set name to this schema builder
        builder.SetSchemaName(schema_name)
        builder.SetDocumentation(schema_documentation)
        
        # run the field builder
        if field_builder:
            field_builder(builder)
            
        schema = builder.Finish()
    return schema


def verify_schema_data_storage_based(
    doc, 
    schema_name,
    schema_docs, 
    schema_guid,
    access_level_read=AccessLevel.Public,
    access_level_write=AccessLevel.Public,
    vendor_id = VENDOR_ID, 
    field_builder=None):
    """
    Verify if the schema exists in the file. if not it will attempt to create it and add to data storage element.
    
   
    :return:
        Result class instance.

        - result.status. True if schema was verified or created successfully, otherwise False.
        - result.message will contain messages status messages.
        - result.result tuple (schema,data_storage)

        On exception:

        - result.status (bool) will be False.
        - result.message will contain generic exception message.
        - result.result will be empty

    :rtype: :class:`.Result`
    """

    return_value = Result()
    
    schema = None
    data_storage = None
    
    try:
        # check if extended storage is set up in the file
        if not does_schema_exist(schema_guid):
            return_value.append_message("Schema does not exists in the file.")
            # set up the schema in the file
            schema = create_schema  (
                schema_name=schema_name, 
                schema_documentation=schema_docs, 
                string_guid=schema_guid,
                vendor_id=vendor_id,
                access_level_read=access_level_read,
                access_level_write=access_level_write,
                field_builder=field_builder
            ) 
            
            # setup the data storage in the file
            data_storage_result = create_project_data_storage(doc, schema)
            if data_storage_result==False:
                message = "Failed to create data storage: {}".format(data_storage_result.message)
                return_value.update_sep(False, message)
                return return_value
            else:
                # get the data storage element
                data_storage = data_storage_result.result[0]
        else:
            return_value.append_message("Schema exists in the file.")
            # get the schema from the file
            schema = get_schema( schema_guid)
            # get the data storage element from the file
            data_storage = find_data_storage(doc,  schema_guid)
            if data_storage == None:
                return_value.append_message("Data storage element not found in the file. Attempting to create it.")
                
                # attempt to create the data storage element
                data_storage_result = create_project_data_storage(doc, schema)
                if data_storage_result==False:
                    message = "Failed to create data storage: {}".format(data_storage_result.message)
                    return_value.update_sep(False, message)
                    return return_value
                else:
                    return_value.append_message("Data storage element created in the file.")
                    # get the data storage element
                    data_storage = data_storage_result.result[0]
            else:
                return_value.append_message("Data storage element found in the file.")
                
        # add the schema and data storage to the return value as a tuple
        return_value.result.append((schema,data_storage))
        
    except Exception as e:
        message = "Failed to set up schema: {}".format(e)
        return_value.update_sep(False, message)
    
    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit family type helper functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from duHast.Utilities.Objects.result import Result
from duHast.Revit.Common.transaction import in_transaction

from Autodesk.Revit.DB import Transaction

# is the default type name in a family when a catalogue file is in use and not types should exist in the family itself
DEFAULT_CATALOGUE_REFERENCE_FAMILY_TYPE = "Refer To Catalog File"

def create_family_type(doc, type_name):
    """
    Creates a new family type in the family document.

    :param doc: The family document
    :type doc: Document
    :param type_name: The name of the family type
    :type type_name: str
    :return: Result object
    :rtype
    """

    return_value = Result()

    try:
        # check this is a family document
        if not doc.IsFamilyDocument:
            return_value.update_sep(False, "Document is not a family document.")
            return return_value
        
        # Get the FamilyManager
        family_manager = doc.FamilyManager

        def action():
            action_return_value = Result()
            try:
                # create new type
                new_fam_type = family_manager.NewType(type_name)
                action_return_value.update_sep(True, "Created new family type: {}".format(type_name))
                action_return_value.result.append(new_fam_type)
            except Exception as e:
                action_return_value.update_sep(False, "Failed to create family type with exception: {}".format(e))
            return action_return_value
        
        tranny = Transaction(doc, "Creating type: {}".format(type_name))
        create_type_result = in_transaction(tranny=tranny, action=action)
        return_value.update(create_type_result)

        return return_value

    except Exception as e:
        return_value.update_sep(False, "Failed to create family type with exception: {}".format(e))
    
    return return_value


def delete_family_type(doc, type_name):
    """
    Deletes a family type from the family document.

    Note: the type will not be deleted if it is the last type in the family!

    :param doc: The family document
    :type doc: Document
    :param type_name: The name of the family type
    :type type_name: str
    :return: Result object
    :rtype:
    """

    return_value = Result()

    try:

        # check this is a family document
        if not doc.IsFamilyDocument:
            return_value.update_sep(False, "Document is not a family document.")
            return return_value
        
        # Get the FamilyManager
        family_manager = doc.FamilyManager

        # get all the types in the family
        familyTypes = family_manager.Types

        # check the number of types in the family
        if familyTypes.Size == 1:
            return_value.append_message("Cannot delete the last type in the family.")
            return return_value

        delete_type_result = None

        # iterate over family types
        familyTypesItor = familyTypes.ForwardIterator()
        familyTypesItor.Reset()
        while (familyTypesItor.MoveNext()):
        
            familyType = familyTypesItor.Current
            if familyType.Name == type_name:
                
                def action():
                    action_return_value = Result()
                    try:
                        # set the current type as the current type so it can be deleted!
                        family_manager.CurrentType = familyType
                        # delete the current type
                        family_manager.DeleteCurrentType()
                        action_return_value.append_message("Deleted family type: {}".format(type_name))
                    except Exception as e:
                        action_return_value.update_sep(False, "Failed to delete family type with exception: {}".format(e))
                    return action_return_value
                
                tranny = Transaction(doc, "Deleting type: {}".format(type_name))
                delete_type_result = in_transaction(tranny=tranny, action=action)
                return_value.update(delete_type_result)

                break
       
        if not delete_type_result:
            return_value.append_message("No matching type {} in family found.".format(type_name))
        elif delete_type_result and not delete_type_result.status:
            return_value.append_message("No matching type {} in family found.".format(type_name))

    except Exception as e:
        return_value.update_sep(False, "Failed to delete family type with exception: {}".format(e))
    
    return return_value


def get_all_family_type_names(doc):
    """
    return all family type names

    :param doc: The family document
    :type doc: Document
    :return: Result object with the list of type names in .result
    :rtype: 
    """

    return_value = Result()
    type_names = []
    try:
        # check this is a family document
        if not doc.IsFamilyDocument:
            return_value.update_sep(False, "Document is not a family document.")
            return return_value
        
        # Get the FamilyManager
        family_manager = doc.FamilyManager

        # get all the types in the family
        familyTypes = family_manager.Types

        # iterate over family types
        familyTypesItor = familyTypes.ForwardIterator()
        familyTypesItor.Reset()
        while (familyTypesItor.MoveNext()):
        
            familyType = familyTypesItor.Current
            type_names.append(familyType.Name)
        
        return_value.update_sep(True, "Got all type names.")
        return_value.result = type_names

    except Exception as e:
        return_value.update_sep(False, "Failed to get all type names with exception: {}".format(e))
    
    return return_value

   


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains the hRevit shared parameter report functionality.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


import Autodesk.Revit.DB as rdb
from duHast.Revit.SharedParameters.shared_parameters import (
    get_all_shared_parameters,
    param_binding_exists,
    param_binding_exists_2023,
)
from duHast.Utilities.utility import encode_ascii
from duHast.Revit.Common import revit_version as rRev


def get_shared_parameter_report_data(doc, revit_file_path):
    """
    Gets shared parameter data ready for being printed to file.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param revit_file_path: The file hostname, which is added to data returned.
    :type revit_file_path: str
    :return: list of list of parameter properties.
    :rtype: list of list of str
    """

    data = []
    paras = get_all_shared_parameters(doc)

    # get revit version
    revit_version = rRev.get_revit_version_number(doc)
    for p in paras:
        if revit_version <= 2022:
            parameter_definition = p.GetDefinition()
            parameter_bindings = []
            # parameter bindings do not exist in a family document
            if doc.IsFamilyDocument == False:
                parameter_bindings = param_binding_exists(
                    doc,
                    rdb.Element.Name.GetValue(p),
                    parameter_definition.ParameterType,
                )
            # just in case parameter name is not unicode
            parameter_name = "unknown"
            try:
                parameter_name = encode_ascii(rdb.Element.Name.GetValue(p))
            except Exception as ex:
                parameter_name = "Exception: {}".format(ex)
            # build data
            data.append(
                [
                    revit_file_path,
                    p.GuidValue.ToString(),
                    str(p.Id.IntegerValue),
                    parameter_name,
                    str(parameter_bindings),
                ]
            )
        else:
            parameter_definition = p.GetDefinition()
            parameter_bindings = []
            data_type_id = parameter_definition.GetDataType()  # forge type id
            # parameter bindings do not exist in a family document
            if doc.IsFamilyDocument == False:
                # uses forge type id
                parameter_bindings = param_binding_exists(
                    doc, rdb.Element.Name.GetValue(p), data_type_id
                )
            # just in case parameter name is not unicode
            parameter_name = "unknown"
            try:
                parameter_name = encode_ascii(rdb.Element.Name.GetValue(p))
            except Exception as ex:
                parameter_name = "Exception: {}".format(ex)
            # build data
            data.append(
                [
                    revit_file_path,
                    p.GuidValue.ToString(),
                    str(p.Id.IntegerValue),
                    parameter_name,
                    ','.join(parameter_bindings),
                ]
            )
    return data



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit API utility functions for Room boundary loops and elements.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#!/usr/bin/python
# -*- coding: utf-8 -*-
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#

from Autodesk.Revit.DB import SpatialElementBoundaryOptions, CurveLoop, Wall, ModelLine


def get_room_segments(room, spat_opts=SpatialElementBoundaryOptions()):
    """
    Get a rooms boundary segments with the default spatial element boundary options.
    :param room: The room to get for
    :type room: Room
    :return: The boundary segments of the room
    :rtype: List[List[BoundarySegment]]
    """
    return room.GetBoundarySegments(spat_opts)


def get_segment_host(rvt_doc, seg):
    """
    Get the element that the room boundary segment belongs to
    :param rvt_doc: The Revit document
    :type rvt_doc: Document
    :param seg: The boundary segment
    :type seg: BoundarySegment
    """

    # TODO: Add support for linked elements
    return rvt_doc.GetElement(seg.ElementId)


def get_curves_as_curve_loop(rm_crvs):
    """
    Get a list of curves as a CurveLoop object.
    :param rm_crvs: The curves to get for
    :type rm_crvs: List[Curve]
    :return: The curves as a CurveLoop object
    :rtype: CurveLoop
    """
    cl = CurveLoop()
    for crv in rm_crvs:
        cl.Append(crv)
    return cl


def get_spatial_element_filter(filter_only):
    """
    Checks if there is a Wall, ModelLine or both filter and returns a list
    of the types to filter for.
    :param filter_only: The host type to filter for
    :type filter_only: Wall, ModelLine or both in a list
    :return: A list of types to filter for
    :rtype: List[Type] or None
    """

    if filter_only:
        if isinstance(filter_only, list):
            return filter_only
        else:
            return [filter_only]
    else:
        return None


def get_all_segs_from_list(segments):
    """
    Get a list of curves that represents the boundary segments of a room.

    Optional parameter to filter the segments by the host type (wall or
    room separation line).
    :param rvt_doc: The Revit document
    :type rvt_doc: Document
    :param segments: The boundary segments of a room
    :type segments: List[List[BoundarySegment]]
    :param filter_only: Optional: The host type to filter for
    :type filter_only: Wall, ModelLine or both in a list
    :return: A list of curves
    :rtype: List[Curve]
    """
    segs = []

    for closed_seg_set in segments:
        for seg in closed_seg_set:
            segs.append(seg)

    return segs


def get_segment_hosts(rvt_doc, segments, filter_only=None):
    """
    Gets the host element of a boundary segment. With option
    to filter by the host type
    :param rvt_doc: The Revit document
    :type rvt_doc: Document
    :param segments: The boundary segments of a room
    :type segments: List[List[BoundarySegment]]
    :param filter_only: Optional: The host type to filter for
    :type filter_only: Wall, ModelLine or both in a list
    :return: A list of hosts
    :rtype: List[Element]
    """

    segment_list = get_all_segs_from_list(segments)

    type_filter_list = get_spatial_element_filter(filter_only)

    hosts = []

    for seg in segment_list:
        host = get_segment_host(rvt_doc, seg)
        if host:
            if type_filter_list is None:
                hosts.append(host)
            else:
                if host.GetType() in type_filter_list:
                    hosts.append(host)

    return hosts


def get_segments_as_curves(rvt_doc, segments, filter_only=None):
    """
    Get the curves that make up the boundary segments of a room.
    :param rvt_doc: The Revit document
    :type rvt_doc: Document
    :param segments: The boundary segments of a room
    :type segments: List[List[BoundarySegment]]
    :param filter_only: Optional: The host type to filter for
    :type filter_only: Wall, ModelLine or both in a list
    :return: A list of curves
    :rtype: List[Curve]
    """
    segment_list = get_all_segs_from_list(segments)

    type_filter_list = get_spatial_element_filter(filter_only)

    curves = []

    for seg in segment_list:
        host = get_segment_host(rvt_doc, seg)
        if host:
            if type_filter_list is None:
                curves.append(seg.GetCurve())
            else:
                if host.GetType() in type_filter_list:
                    curves.append(seg.GetCurve())

    return curves


def get_only_wall_segments_as_curves(rvt_doc, segments):
    """
    Get the wall segments of a room as curves.
    :param rvt_doc: The Revit document
    :type rvt_doc: Document
    :param segments: The boundary segments of a room
    :type segments: List[List[BoundarySegment]]
    :return: A list of curves
    :rtype: List[Curve]
    """
    return get_segments_as_curves(rvt_doc, segments, filter_only=Wall)


def get_only_rm_sep_lines_as_curves(rvt_doc, segments):
    """
    Get the room separation line segments of a room as curves.
    :param rvt_doc: The Revit document
    :type rvt_doc: Document
    :param segments: The boundary segments of a room
    :type segments: List[List[BoundarySegment]]
    :return: A list of curves
    :rtype: List[Curve]
    """
    return get_segments_as_curves(rvt_doc, segments, filter_only=ModelLine)


def get_only_wall_segments_as_walls(rvt_doc, segments):
    """
    Get the wall segments of a room as walls.
    :param rvt_doc: The Revit document
    :type rvt_doc: Document
    :param segments: The boundary segments of a room
    :type segments: List[List[BoundarySegment]]
    :return: A list of walls
    :rtype: List[Wall]
    """
    return get_segment_hosts(rvt_doc, segments, filter_only=Wall)


def get_only_rm_sep_lines_as_model_lines(rvt_doc, segments):
    """
    Get the room separation line segments of a room as model lines.
    :param rvt_doc: The Revit document
    :type rvt_doc: Document
    :param segments: The boundary segments of a room
    :type segments: List[List[BoundarySegment]]
    :return: A list of model lines
    :rtype: List[ModelLine]
    """
    return get_segment_hosts(rvt_doc, segments, filter_only=ModelLine)


def get_rm_segments_centroid(rvt_doc, segments):
    """
    Get the centroid of the room boundary segments
    :param rvt_doc: The Revit document
    :type rvt_doc: Document
    :param segments: The boundary segments of a room
    :type segments: List[List[BoundarySegment]]
    :return: The centroid of the room boundary segments
    :rtype: XYZ
    """
    segs = get_all_segs_from_list(segments)
    pts = [seg.GetCurve().GetEndPoint(0) for seg in segs]

    x = sum([pt.X for pt in pts]) / len(pts)
    y = sum([pt.Y for pt in pts]) / len(pts)
    z = sum([pt.Z for pt in pts]) / len(pts)
    return x, y, z



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Family type data processor class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Family.Data.Objects.ifamily_processor import IFamilyProcessor
from duHast.Revit.Family.Data.Objects.family_type_data_processor_defaults import (
    DATA_TYPE_PROCESSOR as data_type_family_base_processor,
)
from duHast.Revit.Family.Data.Objects import family_type_data as rFamData
from duHast.Revit.Family.family_types_get_data_from_xml import get_type_data_via_XML_from_family_file
from duHast.Utilities.files_io import get_file_name_without_ext
from duHast.Utilities import util_batch_p as uBP
from duHast.Utilities.Objects import result as res
from duHast.Revit.Family.Data.Objects.family_type_data_storage_manager import FamilyTypeDataStorageManager


class FamilyTypeProcessor(IFamilyProcessor):

    data_type = data_type_family_base_processor

    def __init__(
        self,
        session_id=None,
        pre_actions=None,
        post_actions=None,
        family_file_path=None,
        revit_application=None,
    ):
        """
        Class constructor.
        """

        # store data type  in base class
        super(FamilyTypeProcessor, self).__init__(
            data_type=FamilyTypeProcessor.data_type,
            pre_actions=pre_actions,
            post_actions=post_actions,
        )

        if session_id != None:
            self.session_id = uBP.adjust_session_id_for_directory_name(session_id)
        else:
            self.session_id = session_id
        
        # store the file path of the family file
        self.family_file_path = family_file_path

        # store the revit application object
        self.revit_application = revit_application

        # add default pre actions
        if pre_actions == None:
            self.pre_actions=[self._pre_action_get_xml_root]
        else:
            self.pre_actions.append(self._pre_action_get_xml_root)
        
        # currently no post actions...
       

    def process(self, doc, root_path, root_category_path):
        """
        Calls processor instance with the document and root path provided and adds processor instance to class property .data

        :param doc: Current family document.
        :type doc: Autodesk.Revit.DB.Document
        :param rootPath: The path of the nested family in a tree: rootFamilyName :: nestedFamilyNameOne :: nestedFamilyTwo\
            This includes the actual family name as the last node.
        :type rootPath: str
        :param rootCategoryPath: The path of the nested family in in terms of category in a tree: rootFamilyCategory :: nestedFamilyOneCategory :: nestedFamilyTwoCategory\
            This includes the actual family category as the last node.
        :type rootCategoryPath: str
        """

        dummy = rFamData.FamilyTypeData(root_path, root_category_path)
        dummy.process(
            doc,
            self.session_id,
        )
        self.data.append(dummy)
    

    def _pre_action_get_xml_root(self, doc):
        """
        Pre action to get the xml data from the root family.
        
        - need to use the family file on disc to get the xml data...which I can get through the ini...

        :param doc: Current family document.
        :type doc: Autodesk.Revit.DB.Document
        """

        return_value = res.Result()
        try:
            # note to self:
            # this will add an xml family_type_data_storage_manager of the root family to the data property of the processor
            # will need to return a result object confirming success or failure
            family_name = get_file_name_without_ext(self.family_file_path)
            family_category = doc.OwnerFamily.FamilyCategory.Name

            # get the type data from the family file
            type_data_result = get_type_data_via_XML_from_family_file(
                self.revit_application,
                family_name, 
                self.family_file_path, 
                True
            )

            # check if type data was retrieved successfully
            if type_data_result.status and len(type_data_result.result) > 0:
                # in the moment data contains a list of storage objects rather than a single storage manager object
                storage_manager = type_data_result.result[0]

                # data needs to be of type FamilyTypeData...
                dummy = rFamData.FamilyTypeData(family_name, family_category)
                # add the parameter storage to the dummy object
                for type_storage in storage_manager.family_type_data_storage:
                    # we actually want the nested parameter storage here to get a single entry per
                    # family, family type and parameter
                    for  parameter_storage in type_storage.parameters:
                        dummy.add_data(parameter_storage)
                    
                # add the dummy object to the data property of the processor
                self.data.append(dummy)

                return_value.update_sep(True, "Pre Action Get XML Root successful.")
            else:
                return_value.update_sep(False, type_data_result.message)

        except Exception as e:
            return_value.update_sep(
                False,
                "Post Action Update shared parameters data failed with exception: {}".format(
                    e
                ),
            )
        return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to revit revision sequences.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

# import Autodesk
import Autodesk.Revit.DB as rdb
from duHast.Utilities.Objects import result as res
from duHast.Revit.Common import transaction as rTran


def get_revision_seq_of_name(doc, revision_sequence_name):
    """
    Gets a revision sequence by its name. If no match is found None is returned!

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param revision_sequence_name: The sequence name
    :type revision_sequence_name: str

    :return: The matching sequence or None
    :rtype: Autodesk.Revit.DB.RevisionNumberingSequence
    """

    col = rdb.FilteredElementCollector(doc).OfClass(rdb.RevisionNumberingSequence)
    for c in col:
        if c.Name == revision_sequence_name:
            return c
    return None


def create_revision_alpha_seq(
    doc, revision_sequence_name, alpha_settings=rdb.AlphanumericRevisionSettings()
):
    """
    Creates a revision sequence with provided name and settings.

    Will throw an exception if sequence creation failed.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param revision_sequence_name: The name of the sequence
    :type revision_sequence_name: str
    :param alpha_settings: Custom settings, defaults to rdb.AlphanumericRevisionSettings()
    :type alpha_settings: Autodesk.Revit.DB.AlphanumericRevisionSettings, optional

    :return: The new sequence.
    :rtype: Autodesk.Revit.DB.RevisionNumberingSequence
    """

    def action():
        action_return_value = res.Result()
        try:
            seq = rdb.RevisionNumberingSequence.CreateAlphanumericSequence(
                doc, revision_sequence_name, alpha_settings
            )
            action_return_value.update_sep(
                True, "Created sequence: {}".format(revision_sequence_name)
            )
            action_return_value.result.append(seq)
        except Exception as e:
            action_return_value.update_sep(
                False,
                "Failed to create sequence {} with exception: {}".format(
                    revision_sequence_name, e
                ),
            )
        return action_return_value

    transaction = rdb.Transaction(
        doc, "Creating sequence: {}".format(revision_sequence_name)
    )
    transaction_value = rTran.in_transaction(transaction, action)
    if transaction_value.status and len(transaction_value.result) > 0:
        return transaction_value.result[0]
    else:
        raise ValueError(
            "Failed to create sequence: {} with exception:{}".format(
                revision_sequence_name, transaction_value.message
            )
        )



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit rooms. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr

clr.AddReference("System.Core")
from System import Linq

clr.ImportExtensions(Linq)
import System

# import Autodesk
from Autodesk.Revit.DB import (
    BuiltInCategory,
    FilteredElementCollector,
    SpatialElementBoundaryOptions,
)

# --------------------------------------------- utility functions ------------------


def get_all_rooms(doc):
    """
    Gets a list of rooms from the model using built in category.

    Note: This appears to be slightly faster than using a RoomFilter.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: All the rooms in the model as a list.
    :rtype: List Autodesk.Revit.DB.Architecture.Room
    """

    return FilteredElementCollector(doc).OfCategory(BuiltInCategory.OST_Rooms).ToList()


def get_all_rooms_as_collector(doc):
    """
    Gets a list of rooms from the model using built in category.

    Note: This appears to be slightly faster than using a RoomFilter.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: All the rooms in the model as a list.
    :rtype: List Autodesk.Revit.DB.Architecture.Room
    """

    return FilteredElementCollector(doc).OfCategory(BuiltInCategory.OST_Rooms)


def get_unplaced_rooms(doc):
    """
    Gets a list of unplaced rooms from the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: All the unplaced rooms in the model as a list.
    :rtype: List Autodesk.Revit.DB.Architecture.Room
    """

    coll = get_all_rooms_as_collector(doc)
    unplaced = []
    for r in coll:
        if r.Location == None:
            unplaced.append(r)
    return unplaced


def get_not_enclosed_rooms(doc):
    """
    Gets a list of not enclosed rooms from the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: All the unenclosed rooms in the model as a list.
    :rtype: List Autodesk.Revit.DB.Architecture.Room
    """

    coll = get_all_rooms_as_collector(doc)
    unplaced = []
    boundary_option = SpatialElementBoundaryOptions()
    for r in coll:
        boundary_segments = r.GetBoundarySegments(boundary_option)
        if (
            r.Area == 0.0
            and r.Location != None
            and (boundary_segments == None or len(boundary_segments)) == 0
        ):
            unplaced.append(r)
    return unplaced


def get_redundant_rooms(doc):
    """
    Gets a list of redundant rooms from the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: All the redundant rooms in the model as a list.
    :rtype: List Autodesk.Revit.DB.Architecture.Room
    """

    coll = get_all_rooms_as_collector(doc)
    unplaced = []
    boundary_option = SpatialElementBoundaryOptions()
    for r in coll:
        boundary_segments = r.GetBoundarySegments(boundary_option)
        if r.Area == 0.0 and (boundary_segments != None and len(boundary_segments) > 0):
            unplaced.append(r)
    return unplaced


def get_all_placed_rooms(doc):
    """
    Gets a list of all placed rooms from the model.

    Placed criteria is Location is not null and Area > 0.0m2

    :param doc: Revit model document to search.
    :type doc: Autodesk.Revit.DB.Document

    :return: All the placed rooms in the model as a list.
    :rtype: List Autodesk.Revit.DB.Architecture.Room
    """

    all_rooms = get_all_rooms_as_collector(doc)
    return [rm for rm in all_rooms if rm.Location != None and rm.Area > 0.0]



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Varies Element filter / check functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#!/usr/bin/python
# -*- coding: utf-8 -*-
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#


from Autodesk.Revit.DB import Element


def is_element_of_built_in_category(doc, element_id, builtin_categories):
    """
    Checks whether an element is of the built in categories past in.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param element_id: The id of the element to be tested.
    :type element_id: Autodesk.Revit.DB.ElementId
    :param builtin_categories: The builtin category the element does needs to match.
    :type builtin_categories: Autodesk.Revit.DB.Definition
    :return: True if element's builtin category does equals the test category, otherwise False.
    :rtype: bool
    """

    match = False
    el = doc.GetElement(element_id)
    enum_category_id = el.Category.Id.IntegerValue.ToString()
    for bic in builtin_categories:
        if enum_category_id == bic.value__.ToString():
            match = True
            break
    return match


def is_element_not_of_built_in_category(doc, element_id, builtin_categories):
    """
    Checks whether an element is not of the built in categories past in.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param element_id: The id of the element to be tested.
    :type element_id: Autodesk.Revit.DB.ElementId
    :param builtin_categories: The builtin category the element does not needs to match.
    :type builtin_categories: Autodesk.Revit.DB.Definition
    :return: True if element's builtin category does not equals the test category, otherwise False.
    :rtype: bool
    """

    match = True
    el = doc.GetElement(element_id)
    enum_category_id = el.Category.Id.IntegerValue.ToString()
    for bic in builtin_categories:
        if enum_category_id == bic.value__.ToString():
            match = False
            break
    return match


def is_family_name_from_instance(
    doc,
    family_name,  # type: str
    element_id,
):

    """
    Checks whether the family name of a given family instance matches filter value.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param family_name: The string the name of the family needs to match.
    :type family_name: str
    :param element_id: The id of the element to be tested.
    :type element_id: Autodesk.Revit.DB.ElementId
    :return: True if family equals the test string, otherwise False.
    :rtype: bool
    """

    el = doc.GetElement(element_id)
    flag = True
    try:
        if Element.Name.GetValue(el.Symbol.Family) != family_name:
            flag = False
    except Exception:
        flag = False
    return flag


def is_family_name_from_instance_contains(
    doc,
    contains_value,  # type: str
    element_id,
):
    # type: (...) -> bool
    """
    Checks whether the family name of a given family instance contains filter value.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param contains_value: The string the name of the family instance is to be tested for.
    :type contains_value: str
    :param element_id: The id of the element to be tested.
    :type element_id:  Autodesk.Revit.DB.ElementId
    :return: True if family name does contain the test string, otherwise False.
    :rtype: bool
    """

    el = doc.GetElement(element_id)
    flag = True
    try:
        if contains_value not in Element.Name.GetValue(el.Symbol.Family):
            flag = False
    except Exception:
        flag = False
    return flag


def is_family_name_from_instance_does_not_contains(
    doc,
    contains_value,  # type: str
    element_id,
):
    # type: (...) -> bool
    """
    Checks whether the family name of a given family instance does not contains filter value.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param contains_value: The string the name of the family instance is to be tested for.
    :type contains_value: str
    :param element_id: The id of the element to be tested.
    :type element_id: Autodesk.Revit.DB.ElementId
    :return: True if family name does not contain the test string, otherwise False.
    :rtype: bool
    """

    el = doc.GetElement(element_id)
    flag = True
    try:
        if contains_value in Element.Name.GetValue(el.Symbol.Family):
            flag = False
    except Exception:
        flag = False
    return flag


def is_symbol_name_from_instance_contains(
    doc,
    contains_value,  # type: str
    element_id,
):
    # type: (...) -> bool
    """
    Checks whether the family symbol name of a given family instance contains filter value.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param contains_value: The string the name of the family instance is to be tested for.
    :type contains_value: str
    :param element_id: The id of the element to be tested.
    :type element_id: Autodesk.Revit.DB.ElementId
    :return: : True if family name does contain the test string, otherwise False.
    :rtype: bool
    """

    el = doc.GetElement(element_id)
    flag = True
    try:
        if contains_value not in Element.Name.GetValue(el.Symbol):
            flag = False
    except Exception:
        flag = False
    return flag


def is_symbol_name_from_instance_does_not_contains(
    doc,
    contains_value,  # type: str
    element_id,
):
    # type: (...) -> bool

    """
    Checks whether the family symbol name of a given family instance does not contains filter value.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param contains_value: The string the name of the family is to be tested for.
    :type contains_value: string
    :param element_id: The id of the element to be tested.
    :type element_id: Autodesk.Revit.DB.ElementId
    :return: True if symbol name does not contain the test string, otherwise False.
    :rtype: bool
    """

    el = doc.GetElement(element_id)
    flag = True
    try:
        if contains_value in Element.Name.GetValue(el.Symbol):
            flag = False
    except Exception:
        flag = False
    return flag



def get_revit_version_number(doc):
    """
    Returns the revit version as an integer.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: Revit version
    :rtype: int
    """

    app = doc.Application
    return int(app.VersionNumber)
